#!/usr/bin/perl -w
use strict;
use File::Basename;
use Cwd 'abs_path';
use Getopt::Long;
use GO::Parser;
my $docstring=<<EOF
COFACTOR usage:
    ./COFACTOR.pl -libdir lib_dir -datadir data_dir
====================
Mandatory arguments:
====================
    -libdir     I-TASSER suite library directory.
    -datadir    Directory where input structure file "model1.pdb" is located
====================
Optional arguments:
====================
    -homoflag   {real,benchmark}, "real" will use all templates, "benchmark"
                will exclude templates homologous to target by >=0.3
                sequence identity.
    -seqname    A unique name given to each target. If multiple targets are
                run simulanuously, each target should be assigned different
                name. Default value is a random string.
====================
Example:
====================
    Put your target PDB file at $ENV{HOME}/test/model1.pdb. Download 
    I-TASSER library to $ENV{HOME}/ITLIB. Run COFACTOR as:
    ./COFACTOR.pl -libdir $ENV{HOME}/ITLIB -datadir $ENV{HOME}/test
EOF
;

#########Variables to be set dynamically########################################
my $tag        ="";           # tmp folder name
my $datadir    ="";           # target folder
my $run        ="real";       # homoflag
my $model      ="model1.pdb"; # basename of input structure
my $libdir     ="";           # I-TASSER suite library

GetOptions('libdir:s' => \$libdir, 'seqname:s' => \$tag, 
    'model:s' => \$model, 'datadir:s' => \$datadir, 'homoflag:s' => \$run);
die $docstring if (! $datadir || ! $libdir);

$datadir       =abs_path($datadir);
$libdir        =abs_path($libdir);
if (-f $datadir){
    $model     =basename($datadir);
    $datadir   =dirname($datadir);
}
my $s          =basename($datadir);
$tag           ="CF_$s\_".int(rand(1000000)) if (! $tag);
my $modelname="$model";$modelname=~s/\.\S+//;
my $LIBFILE    ="PDBsearchresult_$modelname.dat";
my $OUTPUT_EC  ="ECsearchresult_$modelname.dat";
my $OUTPUT_GO  ="GOsearchresult_$modelname.dat";
my $OUTPUT_BS  ="Bsites_$modelname.dat";
my $OUTPUT_BS2 ="Bpockets_$modelname.dat";

################################################################################
my $start=time();
`hostname`=~/(\S+)/;
my $node=$1;
my $datetime=`date`;
my $path=`pwd`;
print "Hostname: $node\n";
print "Started : $datetime\n";
print "Path    : $path\n";

################################################################################
my $user       ="$ENV{USER}";
my $bindir     =dirname(abs_path(__FILE__)); # COFACTOR bin
my $datdir     ="$libdir/map";               # COFACTOR bin/dat
my $PDB_GOterms="PDB_GOterms.dat";           # PDB GO mapping file
my $PDBlibrary ="$libdir/PDB/";              # dir of I-TASSER PDB database
my $EClibrary  ="$libdir/Enzyme";  # EC library from COFACTOR website
my $GOlibrary  ="$libdir/GO";      # GO library from COFACTOR website
my $BSlibrary  ="$libdir";         # LBS library from COFACTOR website

my $scwrl      ="$bindir/Scwrl4/Scwrl4"; # sidechain packing program
$scwrl         ="$bindir/RASP/RASP" if (!-s $scwrl);
my $nr         ="$libdir/nr/nr";      # the dir I-TASSER nr database
my $blastdir   ="$bindir/blast";      # dir of blast
################################################################################
my $idcut       = "$run";
if ($run eq 'real')
{
    $idcut       = 10.3;
}
elsif($run eq 'benchmark')
{
    $idcut      = 0.30;
}
my $sidcut      = 0.90;
my $FN_cut      = 0.60;
my $TM_cut      = 0.45;
my $distcut     = 8;
my $tancut      = 0.7;
my $BScut       = 0.8;
my $Lweight     = 0.6;

my $PDB_temp    = 30;
my $EC_temp     = 20;
my $GO_temp     = 20;
my $BS_temp     = 75;

my $EXETIME     = 10; #Max time allocated for binding site prediction 
my $message     = '';
my $modeltype   ="native"; 
################################################################################

if(!-s "$datadir/$model")
{
    print "Cannot find target structure at $datadir/$model\n";
    exit;
}


my $workdir  ="/tmp/$user/$tag";
`/bin/mkdir -p $workdir`;
chdir "$workdir";
`/bin/rm -rf $workdir/*`;
my $MTX="$workdir/MTX";
unless(-d "$MTX"){  
    `/bin/mkdir -p  $MTX`;
}
######## Copy executable files #################################################
#Global
`/bin/cp $bindir/TMalign_search2      ./Gsearch1`;
`/bin/cp $bindir/PRTMalign2           ./PRTMalign`;
`/bin/cp $bindir/TMalign2             ./TMalign`;
#Local
`/bin/cp $bindir/EClocal              ./Lsearch1`;
`/bin/cp $bindir/COFACTOR             ./Lsearch2`;

`/bin/cp $bindir/align                ./align`;


`/bin/cp $bindir/get_fasta.pl         ./get_fasta.pl`;
`/bin/cp $bindir/superpose.pl         ./superpose.pl`;

`/bin/cp $datdir/gene_ontology.obo    ./.`;


#my $modelname=$model;$modelname=~s/\.pdb//;
#$model ="$modelname".".pdb";
################################################################################
my %ts=(
        'ALA'=>'A', 'ARG'=>'R', 'ASN'=>'N', 'ASP'=>'D', 'CYS'=>'C', 'GLN'=>'Q', 'GLU'=>'E', 'GLY'=>'G',
        'HIS'=>'H', 'ILE'=>'I', 'LEU'=>'L', 'LYS'=>'K', 'MET'=>'M', 'PHE'=>'F', 'PRO'=>'P', 'SER'=>'S',
        'THR'=>'T', 'TRP'=>'W', 'TYR'=>'Y', 'VAL'=>'V', 'ASX'=>'N', 'GLX'=>'Q', 'UNK'=>'G',
        'A'=>'ALA', 'R'=>'ARG', 'N'=>'ASN', 'D'=>'ASP', 'C'=>'CYS', 'Q'=>'GLN', 'E'=>'GLU', 'G'=>'GLY',
        'H'=>'HIS', 'I'=>'ILE', 'L'=>'LEU', 'K'=>'LYS', 'M'=>'MET', 'F'=>'PHE', 'P'=>'PRO', 'S'=>'SER',
        'T'=>'THR', 'W'=>'TRP', 'Y'=>'TYR', 'V'=>'VAL', 'B'=>'ASX', 'Z'=>'GLX', 'X'=>'UNK');
################################################################################

#####Read NRPDB to PDBALL mapping
my %pdbmap=();
if(!-s "$datdir/NRPDB_PDBALL.map"){$message ="$datdir/NRPDB_PDBALL.map is missing";goto poslast;}
open(HOMOLOGS,"<$datdir/NRPDB_PDBALL.map");
while(my $line=<HOMOLOGS>)
{
    my $center =substr($line,9,8);$center=~s/[\s+|\t]//mg;    
    if($line=~/^\s*$center\s+(\d+)/)
    {
        my $Nhomologs=$1;
        my @T_homologs=split(/\s+/,$line);
        for(my $i=3;$i<=$#T_homologs;$i++)
        {
            push @{$pdbmap{$center}},$T_homologs[$i];
            $i++;
        }
    }
}
close(HOMOLOGS);
    
###### NRPDB-Enzyme library mapping
my %EChomologs=();
if(!-s "$datdir/NRPDB_EC.map"){$message ="$datdir/NRPDB_EC.map is missing";goto poslast;}
open(HOMOLOGS,"<$datdir/NRPDB_EC.map");
while(my $line=<HOMOLOGS>){
    my $center =substr($line,9,8);$center=~s/[\s+|\t]//mg;    
    if($line=~/^\s*$center\s+(\d+)\s+(.*)/){
        my $Nhomologs=$1;
        my @entries=split(/[\s+|\t]/,$2);
        for(my $i=0;$i<=$#entries;$i++){
            push @{$EChomologs{$center}},$entries[$i];
        }
    }
}
close(HOMOLOGS);

###### NRPDB-GO library mapping
my %GOhomologs=();
if(!-s "$datdir/NRPDB_GO.map"){$message ="$datdir/NRPDB_GO.map is missing";goto poslast;}
open(HOMOLOGS,"<$datdir/NRPDB_GO.map");
while(my $line=<HOMOLOGS>){    
    my $center =substr($line,9,8);$center=~s/[\s+|\t]//mg;    
    if($line=~/^\s*$center\s+(\d+)\s+(.*)/){
        my $Nhomologs=$1;
        my @entries=split(/[\s+|\t]/,$2);
        for(my $i=0;$i<=$#entries;$i++){
            push @{$GOhomologs{$center}},$entries[$i];
        }
    }
}
close(HOMOLOGS);
###### GO-Binding Site library mapping
my %GOBShomologs=();
open(HOMOLOGS,"<$datdir/GO_BSITE2.map");
while(my $line=<HOMOLOGS>){
    my $center =substr($line,9,8);$center =~s/\s+//mg;
    if($line=~/^\s*$center\s+(\d+)\s+(.*)/){
        my $Nhomologs=$1;
        my @entries=split(/[\s+|\t]/,$2);
        for(my $i=0;$i<=$#entries;$i++){
            push @{$GOBShomologs{$center}},$entries[$i];
        }
    }
}
close(HOMOLOGS);

###### NRPDB-Binding Site library mapping
my %BShomologs=();
if(!-s "$datdir/NRPDB_BS2.map"){$message ="$datdir/NRPDB_BS2.map is missing";goto poslast;}
open(HOMOLOGS,"<$datdir/NRPDB_BS2.map");
while(my $line=<HOMOLOGS>){
    my $center =substr($line,9,8);$center=~s/[\s+|\t]//mg;    
    if($line=~/^\s*$center\s+(\d+)\s+(.*)/){
        my $Nhomologs=$1;
        my @entries=split(/[\s+|\t]/,$2);
        for(my $i=0;$i<=$#entries;$i++){
            push @{$BShomologs{$center}},$entries[$i];
        }
    }
}
close(HOMOLOGS);

###### Enzyme library 1 (Global)
my %PDBEC=();
if(!-s "$datdir/enzymes.dat"){$message ="$datdir/enzymes.dat is missing";goto poslast;}
open(LIB,"<$datdir/enzymes.dat");
while(my $line=<LIB>){
    chomp($line);
    if($line=~/(\S+)\s+(\S+)/){
	$PDBEC{$1}=$2;
    }  
}
close(LIB);

###### Enzyme library 2 (Local)
my %LEC=();my %LECO=();
open(LIB,"<$datdir/Local_EC.lst");
while(my $line=<LIB>){
    chomp($line);
    if($line=~/(\S+)\s+(\S+)/){
	push @{$LEC{$1}},$line;
	push @{$LECO{$1}},$line;
    }  
}
close(LIB);

###### GO library 1 (Global)
my %PDBGO=();
if(!-s "$datdir/$PDB_GOterms"){$message ="$datdir/$PDB_GOterms is missing";goto poslast;}
open(LIB,"<$datdir/$PDB_GOterms");
while(my $line=<LIB>)
{
    chomp($line);
    if($line=~/(\S+)\s+(\S+)/)
    {   
        $PDBGO{$1}=$2;
    }
}
close(LIB);

###### Information content of GO terms
my %GOcounts=();
open(ALT,"<$datdir/GOcounts.dat");
while(my $line=<ALT>){
    if($line=~/(\S+)\s+(\S+)/){
        $GOcounts{$1}=$2;
    }
}
close(ALT);

###### Gene Ontology file
if(!-s "$datdir/gene_ontology.obo"){$message ="$datdir/gene_ontology.obo is missing";goto poslast;} 
my $parser = new GO::Parser({handler=>'obj'}); # create parser object
$parser->parse("gene_ontology.obo"); # parse file -> objects
my $graph = $parser->handler->graph;

###### Ligand Chemical fingerprint file 
my %ligand_print=();
open(FILE,"<$datdir/ligand2.fpt");
while(my $line=<FILE>)
{
    chomp($line);
    if($line=~/(\S+)\s+(\S+)/)
    {
        my $ligand_name=$1;
        my $fp=$2;
        $ligand_name=~s/\s+//g;
        $fp=~s/\s+//g;
        $ligand_print{$ligand_name}=$fp;
    }
}
close(FILE);

###### Binding Site Library  
my %BSlibrary=();
if(!-s "$datdir/BSITES2.dat"){$message ="$datdir/BSITES2.dat is missing";goto poslast;} 
open(LIB,"<$datdir/BSITES2.dat");
while(my $line=<LIB>)
{
    chomp($line);
    my $pdbname=substr($line,0,5);
    $pdbname   =~s/\s+//;
    push @{$BSlibrary{$pdbname}},$line;    
}
close(LIB);


###### Alterernate IDs for GO terms
my %GOaltids=();
if(!-s "$datdir/GOaltids.dat"){$message ="$datdir/GOaltids.dat is missing";goto poslast;}
open(ALT,"<$datdir/GOaltids.dat");
while(my $line=<ALT>){      
    if($line=~/(\S+)\s+(\S+)/){
	$GOaltids{$1}=$2;
    }
}
close(ALT);

################################################################################
###### STARTING ALL PREDICTIONS FOM HERE
################################################################################
open(ST,">$datadir/status_report");
printf ST "All the library files were read. Starting COFACTOR predictions at %8.3f\n",runhour($start);
close(ST);
######### Copy files from datadir to working directory ##################
if(-s "$datadir/$model")
{
    `/bin/cp $datadir/$model         ./$model`;
}
else
{
    $message ="$model could not be located in $datadir";
    goto poslast;
}

####  Remove any HETATM and renumber PDB file ##################################
my $n=-1000;my $chain='';my $position=0;my %seq=();
my $last_res_id='';my $last_res_name='';
open(OUT,">temp.pdb");
open(FILE,"<$model");
while(my $line=<FILE>)
{    
    if($line=~/^TER/) {goto pdbend;}  # start over if we find a terminus
    if($line=~/^ATOM/)
    {
	my $alt         = substr($line,16,1); #alterate for atom
	if($alt eq " " || $alt eq "A"|| $alt eq "1")
	{
	    $position++;
            if ($n == -1000)
            { 
                $n = 1;# start with the starting number in the chain
                $last_res_id   = substr($line,22,5);
		$last_res_name = substr($line,17,3);
		$seq{$n}=$ts{$last_res_name};	
            }
            my $res_id      = substr($line,22,5);
	    my $res_name    = substr($line,17,3); 
            if(($res_id ne $last_res_id) || ($res_name ne $last_res_name))
            {
                $n++;
                $last_res_id   = $res_id;		
		$last_res_name = $res_name;
		$seq{$n}=$ts{$res_name};		    
            }
            substr($line,6,5)  = sprintf "%5d",$position;
            substr($line,22,5) = sprintf "%4d ",$n;	
	    print OUT "$line";
	}
    }
}
 pdbend:;
close(FILE);
print OUT "TER\n";
close(OUT);
my $Lch=$n;

`/bin/mv temp.pdb $model`;
open(FASTA,">$datadir/seq.txt");
print FASTA ">protein\n";
for(my $i=1;$i<=$Lch;$i++)
{    
    print FASTA "$seq{$i}";
    if(int($i/60)*60==$i)
    {
        print FASTA "\n";
    }
}
print FASTA "\n";
close(FASTA);

#####Generate sequence, run PSI-BLAST and run JSD prediction ######
`/bin/cp $datadir/seq.txt $workdir/.`;


my @seqtxts=();my $sequence1="";my %seqQ=();
@seqtxts=`cat $workdir/seq.txt`;
foreach my $seqtxt(@seqtxts)
{
    goto pos1 if($seqtxt=~/\>/);
    $seqtxt=~s/\s//mg;
    $seqtxt=~s/\n//mg;
    $sequence1=$sequence1.$seqtxt;
  pos1:;
}
my $Lch2=length $sequence1;

if($Lch !=$Lch2)
{
    $message="There is something wrong in between the query PDB file and sequence file. $Lch != $Lch2";
    goto poslast;
}

open(SEQ,">$workdir/protein.seq");
printf SEQ ">protein\n";
for(my $i=1;$i<=$Lch;$i++)
{   
    my $a=substr($sequence1,$i-1,1);
    $seqQ{$i}=$a; 
    printf SEQ "$a";
    if($i==int($i/60)*60)
    {
        printf SEQ "\n";
    }
}
close(SEQ); 

my $blastfile   ="$datadir/blast.out";
my $chk         ="$datadir/psitmp.chk";
my $pssm        ="$datadir/pssm.txt";
my $mtx         ="$datadir/protein.mtx";
if((!-s "$blastfile")||(!-s "$mtx"))
{
    system ("$blastdir/bin/blastpgp  -b 1000 -j 3 -h 0.001 -d $nr -i protein.seq -C psitmp.chk -Q pssm.txt > blast.out");  
    `/bin/cp blast.out  $blastfile`;
    `/bin/cp psitmp.chk $chk`;
    `/bin/cp pssm.txt   $pssm`;
    ########### Make mtx file ############################################################
    `/bin/cp protein.seq psitmp.fasta`;
    `echo psitmp.chk > psitmp.pn`;
    `echo psitmp.fasta > psitmp.sn`;
    system("$blastdir/bin/makemat -P psitmp");
    `/bin/cp psitmp.mtx $mtx`;
    `/bin/cp psitmp.mtx protein.mtx`;     
}
else
{
    `/bin/cp $blastfile $workdir/.`;
    `/bin/cp $mtx       $workdir/.`;
}

if(!-s "$datadir/JSD.dat")
{
    &runJSD($s,$datadir,$user,$blastdir,$nr,$bindir);
    `/bin/cp $datadir/JSD.dat    ./JSD.dat`;
    `/bin/cp $datadir/SOP.dat    ./SOP.dat`;
}
else
{
    `/bin/cp $datadir/JSD.dat    ./JSD.dat`;
    `/bin/cp $datadir/SOP.dat    ./SOP.dat`;
}

if(!-s "$datadir/protein.prf")
{
    &makePRF($s,$datadir,$user,$blastdir,$nr,$bindir);
    `/bin/cp $datadir/protein.prf    ./protein.prf`; 
}
else
{
    `/bin/cp $datadir/protein.prf    ./protein.prf`;
}

open(ST,">$datadir/status_report");
printf ST "Finished running PSI-BLAST and generated JSD and PRF files at %8.3f\n",runhour($start);
printf ST "Now starting with PDB library search.\n";
close(ST);
######### If it is model calculate C-score #####################################
my $MCscore=1; my %cs   =();
if($modeltype eq 'model')
{
    my $modelno= $1 if($model=~/\S+(\d+)\.pdb/);   
    &getCscore($datadir,\%cs);
    $MCscore=($cs{$modelno} + 5)/7;   
    $MCscore= 0.1 if ($MCscore < 0.1);
}
if($MCscore<1)
{
    open(ST,">>$datadir/status_report");
    printf ST "Normalized C-score of model being used is %6.3f\n",$MCscore;
    close(ST);
}
##########Run TM-align search against NR-PDB####################################   
if(-e "$datadir/$LIBFILE"){    
    goto runlibS if(-s "$datadir/$LIBFILE" <100);
    `/bin/cp $datadir/$LIBFILE       ./libsearch.dat`;
}
else{
  runlibS:;
    system("./Gsearch1 $model $PDBlibrary > $workdir/libsearch.dat");    
    `/bin/cp libsearch.dat  $datadir/$LIBFILE`;
}

#########Extend Search against NR-PDB to complete PDB ##########################
my $i=0; my %TNAME_NR=();my %TMscore_NR=();my %TMscore_NR2=(); my %seen_template=();
open(PDBS,"<$workdir/libsearch.dat") || die "Global search result is absent\n";
while(my $line=<PDBS>)
{
    chomp($line);
    if($line=~/Template=\s*(\S+)\s*,\sQuery=\s*\S+\s*,\sLch_t=\s*\d+\,\sLch_q=\s*\d+\,\sTM-score_q=\s*(\S+)\,\sTM-score_t=\s*(\S+)\,\sALN=\s*\d+\,\sSID=\s*\S+/)
    {
	$TNAME_NR{$i}   =$1 ;$TNAME_NR{$i}    =~s/\s+//g;
        $TMscore_NR{$i} =$2 ;$TMscore_NR{$i}  =~s/\s+//g;
	$TMscore_NR2{$i}=$3; $TMscore_NR2{$i} =~s/\s+//g;
	$i++;
    }
}
close(PDBS); 
 PDBsort:;
my @keys1 = sort{$TMscore_NR{$b}<=>$TMscore_NR{$a}} keys %TMscore_NR;
##### Create a local PDBall & copy templates to local
`/bin/mkdir -p $workdir/PDBall`;
my @redundant=(); my $pdb_sel=0;
for(my $j=0;$j<$#keys1;$j++)
{  
    my $seed1 = $TNAME_NR{$keys1[$j]};
    `/bin/cp $PDBlibrary/$seed1\.pdb  $workdir/PDBall/$seed1.pdb`;
    push(@redundant,$seed1);
    if(exists  $pdbmap{$seed1})
    {
        my @ext_templates = @{$pdbmap{$seed1}};
        my %seenpdbid=(); my $pcount=0;
        foreach my $temp(@ext_templates)
        {
            my $ID   =0.00;
            my $pdbid=substr($temp,0,4);
            if(exists $seenpdbid{$pdbid}){next;}
	    my $PDB_file="$PDBlibrary/$temp.pdb";
	    #$PDB_file="$GOlibrary/$temp.pdb"  if(!-s "$PDB_file");
	    #$PDB_file="$EClibrary/$temp.pdb"  if(!-s "$PDB_file");
	    #$PDB_file="$BSlibrary/receptor/$temp.pdb" if(!-s "$PDB_file");
	    next if(!-s "$PDB_file");

            $seenpdbid{$pdbid}=1;
            $pcount++;
            `/bin/cp $PDB_file $workdir/PDBall/$temp\.pdb`;
            if($run ne 'real')
            {
                my $rst=`$workdir/align protein.seq $workdir/PDBall/$temp\.pdb 2`;
                if($rst=~/Identical length:\s*(\d+)/)
                {
                    $ID=$1/$Lch;
                }
            }
            if($ID >= $idcut){next;}
            push(@redundant,$temp);
            $pdb_sel++;
            if($pcount > 10){goto nextseed;}
        }
        if($pdb_sel > $PDB_temp){goto pos2;}
    }
  nextseed:;
}
 pos2:;

my @unique_T = grep { ! $seen_template{$_} ++ } @redundant;
open(ST,">$datadir/status_report");
printf ST "Collected %-5d structural analogs from PDB\n",$#unique_T;
close(ST);
open(LIST,">$workdir/PDBall/list");
print LIST "$#unique_T\n";
foreach my $t(@unique_T)
{
    if(-s "$workdir/PDBall/$t\.pdb")
    {
        print LIST "$t\n";
    }
}
close(LIST);
printf "Running pairwise sequence comparisions between templates\n";
my %ident_pair=();
for(my $i=0;$i<$#unique_T;$i++)
{
    my $t1  =$unique_T[$i];
    my $Lch1=0;
    for(my $j=$i+1;$j<$#unique_T;$j++)
    {
        my $t2  =$unique_T[$j];
        my $Lch2=0;
        my $rst=`./align $workdir/PDBall/$t1\.pdb $workdir/PDBall/$t2\.pdb 1`;
        if($rst=~/Length of sequence 1:\s*(\d+)/)
        {
            $Lch1=$1;
        }
        if($rst=~/Length of sequence 2:\s*(\d+)/)
        {
            $Lch2=$1;
        }
        if($rst=~/Identical length:\s*(\d+)/)
        {
            $ident_pair{$t1,$t2}=$1/$Lch1;
            $ident_pair{$t2,$t1}=$1/$Lch2;
        }
    }
}

##### Run second round of TM-align search for re-ranking
printf "Extending TM-align search to complete PDB\n";
system("./Gsearch1 $model $workdir/PDBall/ > $workdir/PDBsearchresult_ext_$modelname.dat");
my $hits=0; my %Tname_EXT=(); my %ALN_EXT=(); my %RMSD_EXT=();my %TMscore_EXT=();my %IDEN_EXT=();
open(PDBS,"<$workdir/PDBsearchresult_ext_$modelname.dat");
while(my $line=<PDBS>)
{
    chomp($line);
    if($line=~/Template=\s*(\S+)\s*,\sQuery=\s*\S+\s*,\sLch_t=\s*\d+\,\sLch_q=\s*\d+\,\sTM-score_q=\s*(\S+)\,\sTM-score_t=\s*\S+\,\sALN=\s*(\d+)\,\sSID=\s*\S+/)
    {
        $Tname_EXT{$hits}    =$1; $Tname_EXT{$hits}    =~s/\s+//g;
        $TMscore_EXT{$hits}  =$2; $TMscore_EXT{$hits}  =~s/\s+//g;
        $hits++;
    }
}
close(PDBS);


my @keys2 = sort{$TMscore_EXT{$b}<=>$TMscore_EXT{$a}} keys %TMscore_EXT;
my $collected=0;
@redundant=(); %seen_template=();
open(OUT1,">$workdir/similarpdb_$modelname.lst");
for(my $i=0;$i<$hits;$i++)
{
    my $template1         = $Tname_EXT{$keys2[$i]};
    my $rst=`$workdir/TMalign $workdir/PDBall/$template1.pdb $workdir/$model`;
    my $tmscore1=0;my $rmsd1=0; my $sid1=0;my $aln1=0;
    if($rst=~/TM-score=\s*(\S+).*\s+Chain_2/){
        $tmscore1= $1;
    }
    if($rst=~/Aligned length\s*=\s+(\d+),/){
        $aln1 = $1/$Lch;
    }
    if($rst=~/RMSD=\s*(\S+),/){
        $rmsd1= $1;
    }
    if($rst=~/Seq_ID=n_identical\/n_aligned=\s*(\S+)/){
        $sid1=$1;
    }
   
    my $pdbid             = substr($template1,0,5);
    if(exists $seen_template{$template1}){next;}
    if(exists $seen_template{$pdbid}){next;}
    $collected++;  
    $seen_template{$pdbid}=1;
    system("$workdir/superpose.pl $model $workdir/PDBall/$template1\.pdb TMalign $modelname\_$template1\.pdb");
    printf OUT1 "%-8s %-5.4f %-8.4f %-5.4f %-5.4f\n",$template1,$tmscore1,$rmsd1,$sid1,$aln1;
#   if($collected >= 10)
#   {
#        goto TMdone;
#   }
    for(my $j=$i+1;$j<$hits;$j++)
    {
        my $template2         = $Tname_EXT{$keys2[$j]};
        if(($ident_pair{$template1,$template2} >= $sidcut) && ($ident_pair{$template2,$template1} >=$sidcut))
        {
            $seen_template{$template2}=1;            
        }
    }    
}
close(OUT1);
 TMdone:;
print "Collected: $collected\n";
if($collected < 10)
{
    $PDB_temp = $PDB_temp + 25; 
    goto PDBsort;
}

`/bin/cp $workdir/$modelname*.pdb                       $datadir/.`;
`/bin/cp $workdir/similarpdb_$modelname.lst             $datadir/.`;
`/bin/cp $workdir/PDBsearchresult_ext_$modelname.dat    $datadir/.`;

open(ST,">>$datadir/status_report");
printf ST "Selected %-5d structural analogs from PDB\n",$hits;
printf ST "PDB search finished at %8.3f\n",runhour($start);
close(ST);
######## ****** Structural analog detection finished in PDB ****** #############

ECpred:;
print "Starting with EC search\n";
################################################################################
my %template_name=();  my %TMscore=();    my %RMSDa=();  my %IDEN=();  my %Cov   =();
my %FCscoreEC    =();  my %GLscore  =();  my %LCscore=();my %GLscoreR=(); 
my @selected1    =();  my @selected2=();  my %TMtype=(); my %Template_scores=();
%seen_template=();     my $Tcount=0;      
########Extend the search against  1. EC library   #############################
if((-e "$datadir/$OUTPUT_EC")&&(-s "$datadir/$OUTPUT_EC" >100)){goto gopredS;}
my $ALLEC="$workdir/EClib";# Working directory for EC number predictions 
`/bin/mkdir -p $ALLEC`;`/bin/rm -rf $ALLEC/*`;
################################################################################
open(ST,">>$datadir/status_report");
printf ST "Starting with EC number predictions\n";
close(ST);  

for(my $j=0;$j<$#keys1;$j++){
    my $seed1         =$TNAME_NR{$keys1[$j]};
    my $seedtm        =$TMscore_NR{$keys1[$j]};
    $seed1            =~s/\s+//g;
    if(exists $EChomologs{$seed1}){
	my @templates=@{$EChomologs{$seed1}};
	foreach my $template(@templates){
	    my $ID  = 0.0;my $Score1=0.00;
	    if(exists $seen_template{$template}){next;}
	    my $pdb="$EClibrary/$template\.pdb";`/bin/cp $pdb $ALLEC/$template\.pdb`;
	    if($run ne 'real')
	    {
		my $rst=`$workdir/align protein.seq $ALLEC/$template\.pdb 2`;
		if($rst=~/Identical length:\s*(\d+)/){
		    $ID=$1/$Lch;
		}	    
	    }
	    if($ID >= $idcut){next;}
	    if($template ne ''){	  	
	
		my $mtx    ="$EClibrary/MTX/$template\.mtx"; 
		`/bin/cp $mtx $MTX/$template\.mtx`; 
		my $rst2   =`$workdir/PRTMalign $ALLEC/$template\.pdb $workdir/MTX/$template\.mtx $workdir/$model $workdir/protein.prf`;
		$GLscore{$template}= $Score1;
		$Tcount++;					
	    }
	} 
	if($Tcount > $EC_temp*2){goto loopout1;}	
    }
}
 loopout1:;

$Tcount=0;
foreach my $key (sort {$GLscore{$b} <=> $GLscore{$a} } keys %GLscore)
{
    if(exists $seen_template{$key}){next;}
    $Tcount++;
    ##### Top 20 templates from TM-align
    if($Tcount <= $EC_temp)
    {
	$LCscore{$key,1}=0.0; $LCscore{$key,2}='NA';
	push(@selected1,$key);
	$seen_template{$key}=1;
    }   
}
print "Starting reverse\n";
############################################################################################
@keys2=(); $Tcount=0;
@keys2 = reverse sort{$TMscore_NR2{$a} <=> $TMscore_NR2{$b}} keys %TMscore_NR2;
for(my $j=0;$j<$#keys2;$j++){
    my $seed1         =$TNAME_NR{$keys2[$j]};$seed1            =~s/\s+//g;
    my $seedtm        =$TMscore_NR{$keys2[$j]};  
    if(exists $EChomologs{$seed1}){
	my @templates=@{$EChomologs{$seed1}};
	my $TTm=0.00;
	foreach my $template(@templates){
	    my $ID  = 0.0;
	    if(exists $seen_template{$template}){next;}
	    my $pdb="$EClibrary/$template\.pdb";
            $pdb="$PDBlibrary/$template\.pdb" if (!-s $pdb);

            `/bin/cp $pdb $ALLEC/$template\.pdb`;
	    my $rst=`$bindir/align $workdir/protein.seq $ALLEC/$template\.pdb 2`;
	    if($rst=~/Identical length:\s*(\d+)/){
		$ID=$1/$Lch;
	    }
	    if($ID >= $sidcut){next;}
	    if($template ne ''){	  			
		my $mtx="$EClibrary/MTX/$template\.mtx"; 
		`/bin/cp $mtx $MTX/$template\.mtx`;
		my $prf="$EClibrary/PRF/$template\.prf"; 
		`/bin/cp $prf $MTX/$template\.prf`;
                my $Score1=0.00; my $Score2=0.00; my $TM=0.0;
                my $rst2=`$workdir/PRTMalign $ALLEC/$template\.pdb $workdir/MTX/$template\.mtx $workdir/$model $workdir/protein.prf`;
                if($rst2=~/TM-score\s*=(\S+),/){
                    $Score1 = $1;
                }
                my $rst3=`$workdir/PRTMalign  $workdir/$model $workdir/protein.mtx   $ALLEC/$template\.pdb $MTX/$template\.prf`;
                if($rst3=~/TM-score\s*=(\S+),/){
                    $Score2 = $1;
                }
		my $rst4=`$workdir/TMalign $ALLEC/$template\.pdb $workdir/$model `;
		if($rst4=~/TM-score=\s*(\S+).*\s+Chain_2/){
                    $TM = $1;
                }
		if($TM >=0.45)
		{
		    $GLscoreR{$template}= ($Score1 + $Score2)/2;
		}		
		$Tcount++;			
	    }
	}
	if($Tcount >= $EC_temp){goto loopout11;}	
    }
}
 loopout11:;
$Tcount=0;
foreach my $key (sort {$GLscoreR{$b} <=> $GLscoreR{$a} } keys %GLscoreR)
{
    if(exists $seen_template{$key}){next;}
    ##### Only collect reverse hits if approximately of the same fold 
    ##### (single domain templates) which may have been missed above
    if($GLscoreR{$key} >= $FN_cut)
    {
	$LCscore{$key,1}=0.0; $LCscore{$key,2}='NA';
	push(@selected1,$key);
	$GLscore{$key}=$GLscoreR{$key};
    }   
}
############################################################################################
#Check functional promiscuity of fold
my $NEC=0; my %seen_pair=();
for(my $i=0; $i<=10;$i++)
{      
    my $Stemplate1=$selected1[$i];
    my @EC1   =split(/,/,$PDBEC{$Stemplate1}); 
    for (my $j=$i+1;$j<10;$j++)
    {    
	my $Stemplate2=$selected1[$j];
	if(exists $seen_pair{$Stemplate1,$Stemplate2}){next;}
	else{$seen_pair{$Stemplate1,$Stemplate2}=1;$seen_pair{$Stemplate2,$Stemplate1}=1;}
	my @EC2       =split(/,/,$PDBEC{$Stemplate2});
	my $Fsim      =EC_Similarity(\@EC1,\@EC2);
	if($Fsim < 0.75){$NEC++;}	
    }   
}

foreach my $template(@selected1)
{    
    if(exists $LEC{$template})
    {
	push(@selected2,$template);
	$TMtype{$template}='K';
    }
    else
    {
	my @PBS=();
	@PBS=&get_potential_binding($template,$EClibrary,$bindir,$user,$workdir);
	$TMtype{$template}='P';
	foreach my $l(@PBS)
	{			
	    push @{$LEC{$template}},$l;
	}
	if($#PBS>=0)
	{
	    push(@selected2,$template);
	}
    }
}
open(FILE,">$ALLEC/list");
foreach my $name(@selected2)
{
    my @ligN=@{$LEC{$name}};
    foreach my $line(@ligN)
    {
	print FILE "$line\n";
    }       
}
close(FILE);
if(!-s "$datadir/EClocal_$modelname\.dat")
{
    my $rst=`$workdir/Lsearch1 $workdir/JSD.dat $workdir/$model $ALLEC/ > $workdir/EClocal_$modelname\.dat`;
    `/bin/cp $workdir/EClocal_$modelname\.dat $datadir/EClocal_$modelname\.dat`;
}
else
{
    `/bin/cp   $datadir/EClocal_$modelname\.dat   $workdir/EClocal_$modelname\.dat`;
}
my %sT=();
open(LS,"<$workdir/EClocal_$modelname\.dat");
my @arr=<LS>;
close(LS);
my $Lfile=join('',@arr);my @align=split(/\$\$\$/,$Lfile);
foreach my $ali(@align)
{
    my $template=''; my $bscore  =''; my $LTM  =0.0; my $PID  =0.00; my @residues=();my $idali1=0.00; my $idali2=0.00;my $nres=0;
    my @flines=split(/\n/,$ali);
    foreach my $rline(@flines)
    {
	if($rline=~/Template=\s*(\S+)\s+(\S+)\s+(\S+),\s+BS=(\S+),\s+LTM=(\S+),\s+SID=(\S+),\s+COV=(\S+),\s+ERR=\s*(\S+),\s+IDALI1=(\S+),\s+IDALI2=(\S+)/)
	{
	    $template=$1;
	    $LTM     =$5;
	    $PID     =$6;
	    $bscore  =$4;  
	    $idali1  =$9;
	    $idali2  =$10;
	}	   
	elsif($rline=~/\s+\S+\s+(\S+)\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+\*/)
	{
	    $nres++;
	    my $r    =$2;
	    if($1 eq $3)
	    {        
	    	push(@residues,$r);
	    }
	}
    }
    if($template ne '')
    {
	my $idcut  = 0.60; my $BScut = 0.90; my $LTMcut=0.6;
	if($NEC > 15)
	{
	    #### Probable promiscous fold
	    $idcut  = 0.80; $LTMcut= 0.8;
	    #### Lower cutoff for predicted templates but higher BS-score
	    if ($TMtype{$template} eq 'P'){$idcut  = 0.40; $BScut =1.30;$idali2=$PID;}
	}
	elsif($NEC < 15)
	{#### Probable Non-promiscous fold
	     $idcut  = 0.30;$idali2=$PID;
	     if($TMtype{$template} eq 'P'){$BScut=1.0;}	
	}	

	if($bscore >=0.9)
	{
	    $bscore  = $bscore * $idali1;
	}
	else
	{
	    $bscore  = 0;
	}
    }
    if(exists $sT{$template})
    {
	if($bscore >= $LCscore{$template,1})
	{
	    if($LCscore{$template,1} > 0.9)
	    {
		$LCscore{$template,1}=($bscore + $LCscore{$template,1})/2;
#		if($LCscore{$template,2} eq 'NA'){$LCscore{$template,2}='';}
		if($TMtype{$template} ne 'P')
		{
		    $LCscore{$template,2}= $LCscore{$template,2}."\;".join(',',@residues);
		}
	    }
	    elsif($bscore >$LCscore{$template,1})
	    {
		$LCscore{$template,1}=$bscore;
		if($TMtype{$template} ne 'P')
		{
		    $LCscore{$template,2}=join(',',@residues);
		}
	    }
	    $sT{$template}=1;
	}
    }
    else
    {
        $LCscore{$template,1}=$bscore;
        $LCscore{$template,2}=join(',',@residues);
        $sT{$template}=1;
    }   
}

foreach my $name(@selected1)
{
    if(exists $LCscore{$name,1})
    {
	$Template_scores{$name}= $GLscore{$name} + ($Lweight * $LCscore{$name,1});
    }
    else
    {
	$Template_scores{$name}= $GLscore{$name};
    }    
   
}
############################################################
my @ECkeys     = reverse sort{$Template_scores{$a} <=> $Template_scores{$b}} keys %Template_scores;
my $NumMols    = $#ECkeys+1;
for (my $i=0;$i<$NumMols;$i++)
{    
    my $Stemplate    =$ECkeys[$i];
    my @EC1   =split(/,/,$PDBEC{$Stemplate});    		
    my $rst=`$workdir/TMalign $ALLEC/$Stemplate.pdb $workdir/$model`;
    my $TM=0;my $RM=0; my $ID=0;my $ALN=0;
    if($rst=~/TM-score=\s*(\S+).*\s+Chain_2/){
        $TMscore{$Stemplate}= $1;
    }
    if($rst=~/Aligned length\s*=\s+(\d+),/){
	$Cov{$Stemplate}= $1/$Lch;           
    }
    if($rst=~/RMSD=\s*(\S+),/){
	$RMSDa{$Stemplate}= $1;
    }
    if($rst=~/Seq_ID=n_identical\/n_aligned=\s*(\S+)/){
        $IDEN{$Stemplate}=$1;
    }
    my $y                    = $Template_scores{$Stemplate};    
    my $x                    = - $MCscore *($y - $FN_cut);	
    $x                       = -0.1 if($x > - 0.1);
    $FCscoreEC{$Stemplate}   = (2/(1+exp($x)) + 0.01)-1 +(0.01*$GLscore{$Stemplate});    
}

my @ECkeys2     = reverse sort{$FCscoreEC{$a} <=> $FCscoreEC{$b}} keys %FCscoreEC;
open(OUT1,">$workdir/$OUTPUT_EC");
my %seenECNO=();my $predictions=0;
for(my $i=0;$i<$#ECkeys2;$i++)
{    
    my $template = $ECkeys2[$i];
    my $ECNO     = $PDBEC{$template};
    $predictions++;
    my $from ='P';
    $from = 'K' if(exists $LECO{$template});
    $LCscore{$template,2}='NA' if($LCscore{$template,2} eq ''); 
#    if(exists $LCscore{$template,2})
#    {
	printf OUT1 "%-8s %-5.4f %-8.4f %-5.4f %-5.4f %-5.4f %-20s $LCscore{$template,2}\n",$template,$TMscore{$template},$RMSDa{$template},$IDEN{$template},$Cov{$template},$FCscoreEC{$template},$PDBEC{$template};
#    }
#    else
#    {
#	printf OUT1 "%-8s %-5.4f %-8.4f %-5.4f %-5.4f %-5.4f %-20s NA\n",$template,$TMscore{$template},$RMSDa{$template},$IDEN{$template},$Cov{$template},$FCscoreEC{$template},$PDBEC{$template};
#    }    
}
close(OUT1);

`/bin/cp $workdir/$OUTPUT_EC              $datadir/.`;
`/bin/cp $workdir/EClocal_$modelname\.dat $datadir/.`;
#`/bin/rm -rf $ALLEC`;
open(ST,">>$datadir/status_report");
printf ST "EC search finished at %8.3f\n",runhour($start);
close(ST);
print "Finished EC\n";
##########################################################################
 gopredS:;
my %FCscoreGO =(); my %GOscore=(); 
%template_name=(); %TMscore=(); %RMSDa=(); %IDEN=(); %Cov=();
%GLscore  =(); %LCscore=();  %GLscoreR=(); 
@selected1 =();  @selected2=(); %TMtype=(); %Template_scores=();
%seen_template=(); $Tcount=0;

########  ^^^^^^ Extend the search against  2. GO library ^^^^^^   #######
my $OUTPUT_GOMF="$OUTPUT_GO";$OUTPUT_GOMF=~s/\.dat//;$OUTPUT_GOMF="$OUTPUT_GOMF"."_MF.dat";
my $OUTPUT_GOBP="$OUTPUT_GO";$OUTPUT_GOBP=~s/\.dat//;$OUTPUT_GOBP="$OUTPUT_GOBP"."_BP.dat";
my $OUTPUT_GOCC="$OUTPUT_GO";$OUTPUT_GOCC=~s/\.dat//;$OUTPUT_GOCC="$OUTPUT_GOCC"."_CC.dat";
if((-s "$datadir/$OUTPUT_GOMF")&&(-s "$datadir/$OUTPUT_GOBP")&&(-s "$datadir/$OUTPUT_GOCC")){goto bspredS;}

#if((-e "$datadir/$OUTPUT_GO")&&(-s "$datadir/$OUTPUT_GO">100)){goto bspredS;}
my $ALLGO="$workdir/GOlib";# Working directory for GO term predictions 
`/bin/mkdir -p $ALLGO`;`/bin/rm -rf $ALLGO/*`;
########################################################
for(my $j=0;$j<$#keys1;$j++){
    my $seed1         =$TNAME_NR{$keys1[$j]};
    my $seedtm        =$TMscore_NR{$keys1[$j]};
    $seed1            =~s/\s+//g;
    if(exists $GOhomologs{$seed1}){
	my @templates=@{$GOhomologs{$seed1}};
        foreach my $template(@templates){
            my $ID  = 0.0; my $rst='';
            my $pdb="$GOlibrary/$template\.pdb";
            next if (!-s "$pdb"); # chengxin: skip templates not in GO library
            `/bin/cp $pdb $ALLGO/$template\.pdb`;
            if($run ne 'real'){
                $rst=`$bindir/align $workdir/protein.seq $ALLGO/$template\.pdb 2`;
                if($rst=~/Identical length:\s*(\d+)/){
                    $ID=$1/$Lch;
                }
            }
            if($ID >= $sidcut){next;}
            if($template ne ''){
                my $mtx="$GOlibrary/MTX/$template\.mtx"; `/bin/cp $mtx $MTX/$template\.mtx`;
                my $prf="$GOlibrary/PRF/$template\.prf"; `/bin/cp $prf $MTX/$template\.prf`;
                my $Score1=0.00;
                my $rst2=`$workdir/PRTMalign $ALLGO/$template\.pdb $workdir/MTX/$template\.mtx $workdir/$model $workdir/protein.prf`;
                if($rst2=~/TM-score\s*=(\S+),/){
                    $Score1 = $1;
                }		
                $GLscore{$template}= $Score1;
                $Tcount++;
            }
        }
        if($Tcount >= $GO_temp*2){goto loopout2;}
    }
}
 loopout2:;
$Tcount=0;
foreach my $key (sort {$GLscore{$b} <=> $GLscore{$a} } keys %GLscore)
{
    if(exists $seen_template{$key}){next;}
    $Tcount++;
    ##### Top 20 templates 
    if($Tcount <= $EC_temp)
    {
	$LCscore{$key,1}=0.0; $LCscore{$key,2}='NA';
	push(@selected1,$key);
	$seen_template{$key}=1;
    }   
}

@keys2=();$Tcount=0; 
@keys2 = reverse sort{$TMscore_NR2{$a} <=> $TMscore_NR2{$b}} keys %TMscore_NR2;

for(my $j=0;$j<$#keys2;$j++){
    my $seed1         =$TNAME_NR{$keys2[$j]};$seed1            =~s/\s+//g;
    my $seedtm        =$TMscore_NR{$keys2[$j]};
    if(exists $GOhomologs{$seed1}){
	my @templates=@{$GOhomologs{$seed1}};    
	my $TTm=0.00;
	foreach my $template(@templates){
	    my $ID  = 0.0;
	    if(exists $seen_template{$template}){next;}
	    my $pdb="$GOlibrary/$template\.pdb";`/bin/cp $pdb $ALLGO/$template\.pdb`;
	    my $rst=`$bindir/align $workdir/protein.seq $ALLGO/$template\.pdb 2`;
	    if($rst=~/Identical length:\s*(\d+)/){
		$ID=$1/$Lch;
	    }
	    if($ID >= $sidcut){next;}
	    if($template ne ''){	  			
		my $mtx="$GOlibrary/MTX/$template\.mtx"; `/bin/cp $mtx $MTX/$template\.mtx`;
		my $prf="$GOlibrary/PRF/$template\.prf"; `/bin/cp $prf $MTX/$template\.prf`;
                my $Score1=0.00; my $Score2=0.00; my $TM=0.0;
                my $rst2=`$workdir/PRTMalign $ALLGO/$template\.pdb $workdir/MTX/$template\.mtx $workdir/$model $workdir/protein.prf`;
                if($rst2=~/TM-score\s*=(\S+),/){
                    $Score1 = $1;
                }
                my $rst3=`$workdir/PRTMalign  $workdir/$model $workdir/protein.mtx  $ALLGO/$template\.pdb $MTX/$template\.prf`;
                if($rst3=~/TM-score\s*=(\S+),/){
                    $Score2 = $1;
                }
		my $rst4=`$workdir/TMalign $ALLGO/$template\.pdb $workdir/$model `;
		if($rst4=~/TM-score=\s*(\S+).*\s+Chain_2/){
                    $TM = $1;
                }
		if($TM >=0.45)
		{
		    $GLscoreR{$template}= ($Score1 + $Score2)/2;
		}		
		$Tcount++;			
	    }
	}
	if($Tcount >= $GO_temp){goto loopout21;}	
    }
}
 loopout21:;
$Tcount=0;
foreach my $key (sort {$GLscoreR{$b} <=> $GLscoreR{$a} } keys %GLscoreR)
{
    if(exists $seen_template{$key}){next;}
    ##### Only collect reverse hits if approximately of the same fold 
    ##### (single domain templates) which may have been missed above
    if($GLscoreR{$key} >= $FN_cut)
    {
	$LCscore{$key,1}=0.0; $LCscore{$key,2}='NA';
	push(@selected1,$key);
	$GLscore{$key}=$GLscoreR{$key};
    }   
}

##############################################################################
my %parent=();
open(FILE,">$ALLGO/list");
foreach my $template(@selected1)
{
    if(exists $GOBShomologs{$template})
    {
        my @homologs=@{$GOBShomologs{$template}};
        foreach my $hom(@homologs)
        {
            my $pdb="$BSlibrary/receptor/$hom\.pdb";
            $TMtype{$hom}='K';
            `/bin/cp $pdb $ALLGO/$hom\.pdb`;
            my @ligN=@{$BSlibrary{$hom}};my %seen_lig=();
            foreach my $line(@ligN)
            {
                my $lig = substr($line,7,3);$lig =~s/\s+//;
                if(exists $seen_lig{$lig}){next;}
                $seen_lig{$lig}=1;
                print FILE "$line\n";
		
            }
            $parent{$hom}=$template;
        }
    }
    else
    {     
	my @PBS=();
        @PBS=&get_potential_binding($template,$GOlibrary,$bindir,$user,$workdir);
        $TMtype{$template}='P';
	if($#PBS>=0)
        {
	    foreach my $line(@PBS)
	    {
		print FILE "$line\n";
	    }
	}   
        $parent{$template}=$template;
    }
}
close(FILE);
#GO templates
if(!-s "$datadir/GOlocal_$modelname\.dat")
{
    my $rst=`$workdir/Lsearch1 $workdir/JSD.dat $workdir/$model $ALLGO/ > $workdir/GOlocal_$modelname\.dat`;
    `/bin/cp $workdir/GOlocal_$modelname\.dat $datadir/GOlocal_$modelname\.dat`;
}
else
{
    `/bin/cp $datadir/GOlocal_$modelname\.dat $workdir/GOlocal_$modelname\.dat`;
}

@arr=(); %sT=();$Lfile='';@align='';
open(LS,"<$workdir/GOlocal_$modelname\.dat");
@arr=<LS>;
close(LS);
$Lfile=join('',@arr);@align=split(/\$\$\$/,$Lfile);
foreach my $ali(@align)
{
    my $template=''; my $bscore  =''; my $LTM  =0.0; my $PID  =0.00; my @residues=();my $idali1=0.00; my $idali2=0.00;my $nres=0;
    my @flines=split(/\n/,$ali);
    foreach my $rline(@flines)
    {
        if($rline=~/Template=\s*(\S+)\s+(\S+)\s+(\S+),\s+BS=(\S+),\s+LTM=(\S+),\s+SID=(\S+),\s+COV=(\S+),\s+ERR=\s*(\S+),\s+IDALI1=(\S+),\s+IDALI2=(\S+)/)
        {
            $template=$1;
            $LTM     =$5;
            $PID     =$6;
            $bscore  =$4;
            $idali1  =$9;
            $idali2  =$10;
        }
        elsif($rline=~/\s+\S+\s+(\S+)\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+\*/)
        {
            $nres++;
            my $r    =$2;
            if($1 eq $3)
            {
                push(@residues,$r);
            }
        }
    }
    if($template ne '')
    {
        my $icut  = 0.40; my $BScut = 0.90; my $LTMcut=0.5;#$idali2=$PID;
        if($TMtype{$template} eq 'P'){$BScut=1.1;$icut=0.35;}
        if($bscore >=0.9)
        {
            $bscore  = $bscore * $idali1;
        }
        else
        {
            $bscore  = 0;
        }
        my $p = $parent{$template};
        if(exists $LCscore{$p})
        {
            if($bscore > $LCscore{$p})
            {
                $LCscore{$p}=$bscore;
            }
        }
        else
        {
            $LCscore{$p}=$bscore;
        }
    }
}


foreach my $Stemplate(@selected1)
{
    $LCscore{$Stemplate}=0 if(!exists $LCscore{$Stemplate});
    
    $Template_scores{$Stemplate}= $GLscore{$Stemplate} + ($Lweight * $LCscore{$Stemplate});
    my $rst='';
    $rst=`$workdir/TMalign $ALLGO/$Stemplate.pdb $workdir/$model`;
    my $TM=0;my $RM=0; my $ID=0;my $ALN=0;
    if($rst=~/TM-score=\s*(\S+).*\s+Chain_2/){
        $TMscore{$Stemplate}= $1;
    }
    if($rst=~/Aligned length\s*=\s+(\d+),/){
        $Cov{$Stemplate}= $1/$Lch;
    }
    if($rst=~/RMSD=\s*(\S+),/){
        $RMSDa{$Stemplate}= $1;
    }
    if($rst=~/Seq_ID=n_identical\/n_aligned=\s*(\S+)/){
        $IDEN{$Stemplate}=$1;
    }
}

my @GOkeys     = reverse sort{$Template_scores{$a} <=> $Template_scores{$b}} keys %Template_scores;
$NumMols    = $#GOkeys+1;
for (my $i=0;$i<$NumMols;$i++)
{    
    my $Stemplate    =$GOkeys[$i];   	
    my $rst=`$workdir/TMalign $ALLGO/$Stemplate.pdb $workdir/$model`;
    my $TM=0;my $RM=0; my $ID=0;my $ALN=0;
    if($rst=~/TM-score=\s*(\S+).*\s+Chain_2/){
        $TMscore{$Stemplate}= $1;
    }
    if($rst=~/Aligned length\s*=\s+(\d+),/){
	$Cov{$Stemplate}= $1/$Lch;           
    }
    if($rst=~/RMSD=\s*(\S+),/){
	$RMSDa{$Stemplate}= $1;
    }
    if($rst=~/Seq_ID=n_identical\/n_aligned=\s*(\S+)/){
        $IDEN{$Stemplate}=$1;
    } 
    my $y                    = $Template_scores{$Stemplate};    
    my $x                    = - $MCscore *($y - $FN_cut);	
    $x                       = -0.1 if($x > - 0.1);
    $FCscoreGO{$Stemplate}    = (2/(1+exp($x)) + 0.01)-1 +(0.01*$GLscore{$Stemplate});    
}
@GOkeys     =();
@GOkeys     = reverse sort{$FCscoreGO{$a} <=> $FCscoreGO{$b}} keys %FCscoreGO;
open(OUT,">$workdir/$OUTPUT_GO");
for(my $j=0;$j<$#GOkeys;$j++){
    my $template=$GOkeys[$j];   
    printf OUT "%-8s %-5.4f %-8.4f %-5.4f %-5.4f %-5.4f : %-s\n",$template,$TMscore{$template},$RMSDa{$template},$IDEN{$template},$Cov{$template},$FCscoreGO{$template},$PDBGO{$template};   
}   
close(OUT);


my %MF=();my @selectedMF=();
my %BP=();my @selectedBP=();
my %CC=();my @selectedCC=();
my %GO_score=(); my $Nselect=5;
for (my $i=0;$i<$Nselect;$i++)
{    
    my $Stemplate    =$GOkeys[$i];    
    my $score        =$FCscoreGO{$Stemplate};
    my @GOs=split(/,/,$PDBGO{$Stemplate});    
    my %seenterm=();
    foreach my $GO(@GOs){
#       print "STEP0.   $GO  $score\n";
        if($GO=~/^GO/)
        {
            $GO=$GOaltids{$GO}  if(exists $GOaltids{$GO});
            if(exists $GO_score{$GO})
            {
                $GO_score{$GO} *= (1-$score);
            }
            else
            {
                $GO_score{$GO} = (1-$score);
            }
        }
#       print "STEP0.1  $GO  $GO_score{$GO}\n";
    }
}      
my %GO_score2=(); my %scored_term=();

my %root_term=(
		'GO:0003674'=>1,
		'GO:0008150'=>1,
		'GO:0005575'=>1);

foreach my $key (sort {$GO_score{$b} <=> $GO_score{$a} } keys %GO_score)
{
    my $value       = sprintf("%6.3f",$GO_score{$key});
    my $term        = $graph->get_term($key);
    my $value2      = sprintf("%6.3f",(1-$value));
    $GO_score2{$key}= $value2;
    $scored_term{$term}=1;
    my $name      = $term->name;
#   print "STEP:1   $key $GO_score2{$key}  $name\n";
    my $depth =0;
    my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
    while (my $ni = $it->next_node_instance) {
        $depth = $depth + 1;

        my $term2  = $ni->term;
        if($term2 eq $term){next;}
        my $GO_NO = $term2->public_acc;
        my $root='GO:0003674';
        if($term2->get_code_from_namespace eq 'P'){
            $root='GO:0008150';
        }
        if($term2->get_code_from_namespace eq 'C'){
            $root='GO:0005575';
        }

        my $Nc = 1;
        $Nc   = $GOcounts{$GO_NO} if(exists $GOcounts{$GO_NO});
        my $Np   = $GOcounts{$root};

        my $term3 = $graph->get_term($root);

        my $ratio= sprintf("%6.3f",(-log($Nc/$Np)+0.01));
        my $calc_score=0;
        if($ratio >1){$ratio=1;}
        if($value2 ==0){ $calc_score=0;}
        else
        {
            if($term2->get_code_from_namespace eq 'F'){
                $MF{$term2->public_acc} = $value2  + sprintf("%6.3f",($value2*$ratio));
                $calc_score=$MF{$term2->public_acc};
            }
            if($term2->get_code_from_namespace eq 'P'){
                $BP{$term2->public_acc} = $value2  + sprintf("%6.3f",($value2*$ratio));
                $calc_score=$BP{$term2->public_acc};
            }
            if($term2->get_code_from_namespace eq 'C'){
                $CC{$term2->public_acc} = $value2  + sprintf("%6.3f",($value2*$ratio));
                $calc_score=$CC{$term2->public_acc};
            }
#           printf "%s  Term = %8s depth=%d   Calcscore:$calc_score $ratio  %s\n",$term2->name,$term2->public_acc, $depth, $term2->get_code_from_namespace;
        }
        if((exists $GO_score2{$GO_NO}) && ($calc_score >$GO_score2{$GO_NO})){$GO_score2{$GO_NO}=sprintf("%6.3f",$calc_score);}
        else{$GO_score2{$GO_NO}=sprintf("%6.3f",$calc_score);}
#       print "STEP:1.2 $GO_NO $value2  $GO_score2{$GO_NO}  $depth\n";
#       print "$key  $name  Value $GO_NO $calc_score\n";
    }
}
my %GO_depth=();
foreach my $key (sort {$GO_score2{$b} <=> $GO_score2{$a} } keys %GO_score2)
{
    my $term        = $graph->get_term($key);
    my $name        = $term->name;
    my $depth       =0;
    my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
    while (my $ni = $it->next_node_instance) {
        $depth = $depth + 1;
    }
    $GO_depth{$key}=$depth;
#    print "$key $GO_score2{$key} $GO_depth{$key} $name\n";
}
#Molecular function terms
my %seenMF=();my %selectedMF=(); my $MFcount=0;
#my $OUTPUT_GOMF=$OUTPUT_GO;$OUTPUT_GOMF=~s/\.dat//;$OUTPUT_GOMF= "$OUTPUT_GOMF"."_MF.dat";
foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
{
    my $term      = $graph->get_term($key);
    if(exists $root_term{$key}){next;}
    if($term->get_code_from_namespace eq 'F')
    {
#       print "$key $GO_depth{$key} $GO_score2{$key}\n";
        if(exists $seenMF{$key}){next;}
        if($GO_score2{$key} > 0.30)
        {
            $selectedMF{$key}=1;$MFcount++;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenMF{$GO_NO2}=1;
            }
        }
    }
}
if($MFcount ==0)
{
    foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
    {
        my $term      = $graph->get_term($key);
        if(exists $root_term{$key}){next;}
        if($term->get_code_from_namespace eq 'F')
        {
            if(exists $seenMF{$key}){next;}
            $selectedMF{$key}=1;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenMF{$GO_NO2}=1;
            }
        }
    }
}
open(OUT,">$workdir/$OUTPUT_GOMF");
foreach my $key (sort {$GO_score2{$b} <=> $GO_score2{$a} } keys %GO_score2)
{
    if(exists $selectedMF{$key})
    {
        my $term      = $graph->get_term($key);
        my $score     = $GO_score2{$key};  $score   = 1 if ($score > 1);
        printf OUT "%s %3.2f %s \n",$key,$score,$term->name;
    }
}
close(OUT);

#Biological process terms
my %seenBP=();my %selectedBP=();my $BPcount=0;
#my $OUTPUT_GOBP=$OUTPUT_GO;$OUTPUT_GOBP=~s/\.dat//;$OUTPUT_GOBP= "$OUTPUT_GOBP"."_BP.dat";
foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
{
    my $term      = $graph->get_term($key);
    if(exists $root_term{$key}){next;}
    if($term->get_code_from_namespace eq 'P')
    {
        if(exists $seenBP{$key}){next;}
        if($GO_score2{$key} > 0.30)
        {
            $selectedBP{$key}=1;$BPcount++;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenBP{$GO_NO2}=1;
            }
        }
    }
}
if($BPcount ==0)
{
    foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
    {
        my $term      = $graph->get_term($key);
        if(exists $root_term{$key}){next;}
        if($term->get_code_from_namespace eq 'P')
        {

            if(exists $seenBP{$key}){next;}
            $selectedBP{$key}=1;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenBP{$GO_NO2}=1;
            }
        }
    }
}
open(OUT,">$workdir/$OUTPUT_GOBP");
foreach my $key (sort {$GO_score2{$b} <=> $GO_score2{$a} } keys %GO_score2)
{
    if(exists $selectedBP{$key})
    {
        my $term      = $graph->get_term($key);
        my $score     = $GO_score2{$key};  $score   = 1 if ($score > 1);
        printf OUT "%s %3.2f %s \n",$key,$score,$term->name;
    }
}
close(OUT);



#Cellular component terms
my %seenCC=();my %selectedCC=();my $CCcount=0;
#my $OUTPUT_GOCC=$OUTPUT_GO;$OUTPUT_GOCC=~s/\.dat//;$OUTPUT_GOCC= "$OUTPUT_GOCC"."_CC.dat";
foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
{
    my $term      = $graph->get_term($key);
    if(exists $root_term{$key}){next;}
    if($term->get_code_from_namespace eq 'C')
    {
        if(exists $seenCC{$key}){next;}
        if($GO_score2{$key} > 0.30)
        {
            $selectedCC{$key}=1;$CCcount++;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenCC{$GO_NO2}=1;
            }
        }
    }
}
if($CCcount ==0)
{
    foreach my $key (sort {$GO_depth{$b} <=> $GO_depth{$a} } keys %GO_depth)
    {
        my $term      = $graph->get_term($key);
        if(exists $root_term{$key}){next;}
        if($term->get_code_from_namespace eq 'C')
        {
            if(exists $seenCC{$key}){next;}
            $selectedCC{$key}=1;
            my $it = $graph->create_iterator({acc=>"$key",direction=>"up"});
            while (my $ni = $it->next_node_instance)
            {
                my $term2    = $ni->term;
                my $GO_NO2   = $term2->public_acc;
                $seenCC{$GO_NO2}=1;
            }
        }
    }
}
open(OUT,">$workdir/$OUTPUT_GOCC");
foreach my $key (sort {$GO_score2{$b} <=> $GO_score2{$a} } keys %GO_score2)
{
    if(exists $selectedCC{$key})
    {
        my $term      = $graph->get_term($key);
        my $score     = $GO_score2{$key};  $score   = 1 if ($score > 1);
        printf OUT "%s %3.2f %s \n",$key,$score,$term->name;
    }
}
close(OUT);

`/bin/cp $workdir/$OUTPUT_GO             $datadir/.`;
`/bin/cp $workdir/$OUTPUT_GOMF           $datadir/.`;
`/bin/cp $workdir/$OUTPUT_GOBP           $datadir/.`;
`/bin/cp $workdir/$OUTPUT_GOCC           $datadir/.`;
#`rm -rf $ALLGO`;
open(ST,">>$datadir/status_report");
printf ST "GO prediction finished at %8.3f\n",runhour($start);
close(ST);
##########################################################################
 bspredS:;
chdir "$workdir";
########  ^^^^^^ Extend the search against  3. Binding Site library ^^^^^^   #######
my $BSstart=runhour($start);
#Working directories for Binding Site prediction 
my $ABS="$workdir/BSITES";
my $GBS="$workdir/BSITES/GBS";
my $COM="$workdir/BSITES/COMPLEX"; 
my $ALL="$workdir/all";
my $REC="$workdir/receptor";
my $LIG="$workdir/ligand";
my $BSrecorddir="$datadir/BSITE_$modelname";
unless(-d "$BSrecorddir")
{
    `/bin/mkdir -p  $BSrecorddir`;
}

`/bin/mkdir -p  $ABS`;`rm -rf $ABS/*`;
`/bin/mkdir -p  $GBS`;`rm -rf $GBS/*`;  
`/bin/mkdir -p  $COM`;`rm -rf $COM/*`; 
`/bin/mkdir -p  $ALL`;`rm -rf $ALL/*`;
`/bin/mkdir -p  $REC`;`rm -rf $REC/*`;
`/bin/mkdir -p  $LIG`;`rm -rf $LIG/*`;
my %bions=('K'=>1,'MG'=>1,'ZN'=>1,'MN'=>1,'CO'=>1,'FE'=>1,'FE2'=>1,'NI'=>1,'MO'=>1,'CA'=>1,'NA'=>1,'CU1'=>1,'CU2'=>1,'CU'=>1,);  


%template_name=(); %TMscore=(); %RMSDa=(); %IDEN=(); %Cov=(); %seen_template=(); 
@selected1=(); @selected2=();
$Tcount =0; 
for(my $j=0;$j<$#keys1;$j++){
    my $seed1         =$TNAME_NR{$keys1[$j]};
    my $seedtm        =$TMscore_NR{$keys1[$j]};
    $seed1            =~s/\s+//g;
    if(exists $BShomologs{$seed1}){
	my @BStemplates=@{$BShomologs{$seed1}};
	foreach my $template(@BStemplates)
	{
	    my $ID  = 0.0;
	    if(exists $seen_template{$template}){next;}	   
	    my $pdb="$BSlibrary/receptor/$template\.pdb";
	    `/bin/cp $pdb $ALL/$template\.pdb`;
	    if($run ne 'real')
	    {
		my $rst=`$workdir/align protein.seq $ALL/$template\.pdb 2`;
		if($rst=~/Identical length:\s*(\d+)/){
		    $ID=$1/$Lch;
		}	    
	    }
	    if($ID >= $idcut){next;}
	    if($template ne ''){
		`/bin/cp $ALL/$template\.pdb $REC/$template\.pdb`;	
		my $rst2='';
		$rst2=`$workdir/TMalign $ALL/$template\.pdb $model`;
		if($rst2=~/TM-score=\s*(\S+).*\s+Chain_2/){
        		$TMscore{$template} = $1;
    		}
		if($rst2=~/Aligned length\s*=\s+(\d+),/){
		    $Cov{$template} = sprintf("%5.4f",($1/$Lch));		 
		}
		if($rst2=~/RMSD=\s*(\S+),/){
		    $RMSDa{$template} = $1;
		}
		if($rst2=~/Seq_ID=n_identical\/n_aligned=\s*(\S+)/){
        	   $IDEN{$template} = $1;
    		}
		push(@selected1,$template);
		my @ligN=@{$BSlibrary{$template}};
		$Tcount += ($#ligN + 1);
		$seen_template{$template}=1;
		`/bin/cp $BSlibrary/ligand/$template\_* $LIG/.`;
	    }
	} 
	if($Tcount > $BS_temp){goto loopout3;}	
    }
}
 loopout3:;
######Make Process list

open(OUT1,">$workdir/process.lst");
foreach my $template(@selected1)
{
    if(exists $BSlibrary{$template})
    {
       my @ligN=@{$BSlibrary{$template}};
       foreach my $line(@ligN)
       {
	   print OUT1 "$line\n";
       }
    }
}
close(OUT1);
my %profile=();
open(PROFILE,"<$workdir/JSD.dat") || print "ERROR: JSD.dat is missing\n";
while(my $line=<PROFILE>)
{
    if($line=~/(\S+)\s+(\S+)/)
    {
        $profile{$1}=$2;        
    }
}
close(PROFILE);
my $Pcount = 0;my $TNBS=0;
open(RECORD,">$workdir/record.dat");
open(RESULT,">$COM/unsorted.lst");
open(PROCESS,"<$workdir/process.lst");
while(my $line=<PROCESS>)
{
    chomp($line);
    my $template = ''; my $bsno     = '';my  $no_bsres = '';my $bindingr = '';
    if($line=~/(\S+)\s+\:\S+\:(\S+)\:(\d+)\:(\S+)/)
    {
        $template = $1;
        $bsno     = $2;
        $no_bsres = sprintf("%d",$3);
        $bindingr = $4;
    }
    elsif($line=~/(\S+)\s+\: \S+\:(\S+)\:(\d+)\:(\S+)/)
    {
        $template = $1;
        $bsno     = $2;
        $no_bsres = sprintf("%d",$3);
        $bindingr = $4;
    }
    elsif($line=~/(\S+)\s+\:\S+\s+\:(\S+)\:(\d+)\:(\S+)/)
    {
        $template = $1;
        $bsno     = $2;
        $no_bsres = sprintf("%d",$3);
        $bindingr = $4;
    }
    elsif($line=~/(\S+)\s+\:  \S+\:(\S+)\:(\d+)\:(\S+)/)
    {
        $template = $1;
        $bsno     = $2;
        $no_bsres = sprintf("%d",$3);
        $bindingr = $4;
    }    
    else{print "$line dosent matches\n";goto nexttemplate;}
    $Pcount++;
    my @TP_bsres=split(/,/,$bindingr);
    $TNBS += $#TP_bsres;
   
    my $structure1="$workdir/$model";
    my $structure2="$workdir/receptor/$template.pdb"; 
    my %PBS=();	
    &getBinding(\%PBS,$structure1, $structure2,$workdir,$datadir,$Lch,$bindingr,\%profile,$blastdir,$MCscore);
    my $profile_file= "$GBS/$template\_$bsno.dat";
    my $pairs=0;
    open(GBSR,">$profile_file");
    for(my $i=1;$i<=$Lch;$i++)
    {
	if($PBS{$i} > 0){$pairs++;}
	printf GBSR "%-4d  %5.3f\n",$i,$PBS{$i};
    }
    close(GBSR);
  
    open(PROC,">$workdir/list");
    print PROC "$line\n";
    close(PROC);
    my @rst=(); my @residues=();
    my $ligand  ='';my $bscore=0;my $rmsd=0;my $sid=0;my $coverage=0;my $disterr=0;my $idali1=0;my $idali2=0;
    if(!-s "$workdir/ligand/$template\_$bsno.pdb")
    {
	#print "$workdir/ligand/$template\_$bsno.pdb missed\n";
    }
    else
    {
	@rst=`./Lsearch2 $profile_file $model $workdir -c `;
    }

    foreach my $rline(@rst)
    {
	if($rline=~/Template=\s*(\S+)\s+(\S+)\s+(\S+),\s+BS=(\S+),\s+LTM=(\S+),\s+SID=(\S+),\s+COV=(\S+),\s+ERR=\s*(\S+),\s+IDALI1=(\S+),\s+IDALI2=(\S+)/)
	{
	    $template=$1;
	    my $pdbid   =substr($template,0,4);$pdbid   =lc($pdbid);
	    my $pdbchain=substr($template,4,1);$pdbchain=uc($pdbchain);
	    $ligand  =$2;
	    $bsno    =$3;
	    $bscore  =$4;
	    $rmsd    =$5;
	    $sid     =$6;
	    $coverage=$7;
	    $disterr =$8;
	    $idali1  =$9;
	    $idali2  =$10;
	}
	elsif($rline=~/\s*\d+\s+\S+\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\*/)
	{
	    my $r    =$1;
	    push(@residues,$r);
	}
	else
	{
	    next;
	}
    }
    if($#residues>1)
    {
#	printf "%-5s %5.4f %5.4f %5.4f %5.4f %3s %4s %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f :@residues", $template, $TMscore{$template}, $RMSDa{$template}, $IDEN{$template}, $Cov{$template}, $ligand, $bsno, $bscore, $rmsd, $sid, $coverage, $disterr, $idali1,$idali2;
	printf RESULT "%-5s %5.4f %5.4f %5.4f %5.4f %3s %4s %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f :", $template, $TMscore{$template}, $RMSDa{$template}, $IDEN{$template}, $Cov{$template}, $ligand, $bsno, $bscore, $rmsd, $sid, $coverage, $disterr, $idali1,$idali2;  
	printf RESULT join(',',@residues);
	printf RESULT "\n";
	`$scwrl -i $workdir/$model -o temp.pdb -f BSITE_$template\_$bsno.pdb`;	   
        if(-e "temp.pdb")
        {
	    if(-s "temp.pdb" > 10)
	    {
		`/bin/cat temp.pdb BSITE_$template\_$bsno.pdb > $COM/$modelname\_$template\_$bsno.pdb`;
	    }
	    else
	    {
		`/bin/cat $model BSITE_$template\_$bsno.pdb > $COM/$modelname\_$template\_$bsno.pdb`;
	    }
	    `rm temp.pdb`;
	}
	else
        {
		`/bin/cat $model BSITE_$template\_$bsno.pdb > $COM/$modelname\_$template\_$bsno.pdb`;
        }
    }
    print RECORD "@rst\n";
    my $runtime = runhour($start);
    if(($EXETIME - ($runtime-$BSstart)) <= 0.2){goto finishs;}
  nexttemplate:;
}
finishs:;
close(PROCESS);
close(RECORD);
close(RESULT);

`/bin/cp -r process.lst $BSrecorddir/.`;
`/bin/cp -r record.dat  $BSrecorddir/.`;
####Cluster the predicted binding site ligand
 reclust:;
%TMscore =(); %RMSDa  =(); %IDEN =(); %Cov   =(); 
my %hits    =(); my %ligands=(); my %ligand_name=(); my %PC =(); 
my %BS_score=(); my %LTM    =(); my %BS_ID=(); my %BS_cov=(); my %BS_err=(); my %BS_ID1=(); my %BS_ID2=(); my %PBS=();

$i=0; 
my $Ncluster=0;my %Rank_scores=();
open(RESULT,"<$COM/unsorted.lst");
while(my $line2=<RESULT>)
{
    if($line2=~/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\:(\S+)/)
    {
	$hits{$i}	 =$1;
	$TMscore{$i}	 =$2;
	$RMSDa{$i}	 =$3;
	$IDEN{$i}	 =$4; 
	$Cov{$i}	 =$5;
	$ligand_name{$i} =$6;
	$ligands{$i}	 =$7;
	$BS_score{$i}	 =$8;
	$LTM{$i}	 =$9;
	$BS_ID{$i}	 =$10;
	$BS_cov{$i}	 =$11;
	$BS_err{$i}	 =$12;
	$BS_ID1{$i}	 =$13;
	$BS_ID2{$i}	 =$14;
	$PBS{$i}	 =$15;	    
	#Filters
        if((exists $bions{$ligand_name{$i}})&& (($BS_ID1{$i} < 0.50) && ($BS_ID2{$i} < 0.40))){next;}
	if($BS_ID2{$i}   < 0.1)   {next;}
        if($BS_score{$i} < $BScut){next;}
        $Rank_scores{$i} =  $TMscore{$i} + 2.5*$IDEN{$i} + (0.25* $BS_score{$i});
	my $xcom=0;my $ycom=0;my $zcom=0; my $natoms=0;
	open(FILE,"<$COM/$modelname\_$hits{$i}\_$ligands{$i}\.pdb");
	while(my $l=<FILE>)
	{
	    if($l=~/^HETATM/)
	    {
		$natoms++;
		$xcom +=substr($l,30,8);
		$ycom +=substr($l,38,8);
		$zcom +=substr($l,46,8);
	    }		
	}
	close(FILE);
	$PC{$i,1} = $xcom/$natoms;
	$PC{$i,2} = $ycom/$natoms;
	$PC{$i,3} = $zcom/$natoms;	    
	$i++;
    }	
}
close(RESULT);
my $NBST=$i;
if(($NBST < 5) && ($BScut > 0.5)){$BScut =  $BScut - 0.1;goto reclust;}
if(($NBST < 1) && ($BScut > 0.1)){$BScut =  0.1;         goto reclust;}
if($NBST < 1){goto nobs;}
############# Perform Clustering based on Euclidean distance of Binding Site Center and Tannimoto Coeff of ligands###
my %ED=(); my %TN=(); my %Lname=();
for(my $i=0;$i<$NBST;$i++)
{
    $ED{$i,$i}=0;
    $TN{$i,$i}=1;
    my $ligname1=$hits{$i}."_".$ligands{$i};
    $Lname{$i}=$ligand_name{$i};
    for(my $j=$i+1;$j<$NBST;$j++)
    {
        if($i == $j){next;}
        $ED{$j,$j}=0;
        $TN{$j,$j}=1;
        my $dist =sqrt(($PC{$i,1}-$PC{$j,1})**2+
                       ($PC{$i,2}-$PC{$j,2})**2+
                       ($PC{$i,3}-$PC{$j,3})**2);
        $ED{$i,$j}=sprintf("%6.3f",$dist);
        $ED{$j,$i}=sprintf("%6.3f",$dist);
        my $ligname2=$hits{$j}."_".$ligands{$j};
        my $tani = 0.0;
        $tani =&tannimoto($ligand_print{$ligname1},$ligand_print{$ligname2});
        if((exists $bions{$ligand_name{$i}}) && (exists $bions{$ligand_name{$j}}))
        {
            $tani = 1;
        }
        elsif(($ligand_name{$i} eq 'DNA') && ($ligand_name{$j} eq 'DNA'))
        {
            $tani = 1;
        }
        elsif(($ligand_name{$i} eq 'RNA') && ($ligand_name{$j} eq 'RNA'))
        {
            $tani = 1;
        }
        $TN{$i,$j}=$tani;
        $TN{$j,$i}=$tani;
    }
}

my $prediction_no=0; my %clusters=();my %pockets=(); my %Cscore=();

%clusters=&cluster($BSrecorddir,\$Ncluster,$distcut,$tancut,$NBST,\%ED,\%TN,\%Rank_scores,\%BS_score,\%BS_ID1,\%PBS,\%Lname,\%pockets,\%PC,\%Cscore);
 nobs:;
open(RESULT2,">$BSrecorddir/$OUTPUT_BS2");
open(RESULT1,">$BSrecorddir/$OUTPUT_BS");
if($Ncluster >=1)
{
    foreach my $rank (sort {$Cscore{$b} <=> $Cscore{$a} } keys %Cscore)
    {
        my $index = $clusters{$rank};
        $prediction_no++;
        if($prediction_no > 15){next;}
        printf RESULT2 "POCKET %-2d %5.3f %8.3f %8.3f %8.3f\n",$prediction_no,$Cscore{$rank},$pockets{$rank,1},$pockets{$rank,2},$pockets{$rank,3};
	printf RESULT1 "%-4d %5.3f %-5s %5.4f %5.4f %5.4f %5.4f %3s %4s %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f %5.4f:$PBS{$index}\n",$prediction_no,$Cscore{$rank},$hits{$index}, $TMscore{$index}, $RMSDa{$index}, $IDEN{$index}, $Cov{$index}, $ligand_name{$index},$ligands{$index} ,$BS_score{$index}, $LTM{$index},$BS_ID{$index},$BS_cov{$index},$BS_err{$index},$BS_ID1{$index},$BS_ID2{$index};
	`/bin/cp $COM/$modelname\_$hits{$index}\_$ligands{$index}.pdb $BSrecorddir/complex$prediction_no\.pdb`;	
    }
}  
else
{
    printf RESULT1 "No Binding sites was predicted\n";
    printf RESULT2 "No Binding pocket was predicted\n";
}
close(RESULT1);
close(RESULT2);

open(ST,">>$datadir/status_report");
printf ST "Binding site prediction finished at %8.3f\n",runhour($start);
close(ST);



$message = "Cheers!! Everything finished successfully.";


 poslast:;
print "$message\n";
################# endding procedure ######################
my $time=`date`;
printf "ending time: $time";
`sync`;
sleep(1);
`rm -fr $workdir`;
exit();

###Similarity of EC number between the templates
sub EC_Similarity
{
    my ($ec1_ref,$ec2_ref)=@_;
    my @ecs1  =@$ec1_ref;
    my @ecs2  =@$ec2_ref;
    my $Max_score=0;
    foreach my $ec1(@ecs1)
    {
        my @no1=split(/\./,$ec1);
        if($ec1 eq ''){next;}
        foreach my $ec2(@ecs2)
        {
            my @no2=split(/\./,$ec2);
            if($ec2 eq ''){next;}
            my $score =0;
            if($ec2 eq 'NA'){$score =0;next;}
            $score++ if($no1[0] eq $no2[0]);
            $score++ if(($no1[0] eq $no2[0]) &&($no1[1] eq $no2[1]));
            $score++ if(($no1[0] eq $no2[0]) &&($no1[1] eq $no2[1]) && ($no1[2] eq $no2[2]));
            $score++ if(($no1[0] eq $no2[0]) &&($no1[1] eq $no2[1]) && ($no1[2] eq $no2[2]) && ($no1[3] eq $no2[3]));
            $score++ if(($score >=2)&&($no1[3] ne $no2[3]) && (($no1[3] eq '_') || ($no2[3] eq '_')));
            $Max_score= $score if($score >= $Max_score);
        }
    }
    $Max_score=$Max_score/4;
    return ($Max_score);
}

### Subroutine to get potential binding sites in template protein
sub get_potential_binding
{
    my ($protname,$library,$bin,$user,$dir)=@_;

    my %ts=(
            'ALA'=>'A', 'ARG'=>'R', 'ASN'=>'N', 'ASP'=>'D', 'CYS'=>'C', 'GLN'=>'Q', 'GLU'=>'E', 'GLY'=>'G',
            'HIS'=>'H', 'ILE'=>'I', 'LEU'=>'L', 'LYS'=>'K', 'MET'=>'M', 'PHE'=>'F', 'PRO'=>'P', 'SER'=>'S',
            'THR'=>'T', 'TRP'=>'W', 'TYR'=>'Y', 'VAL'=>'V', 'ASX'=>'N', 'GLX'=>'Q', 'UNK'=>'G',
            'A'=>'ALA', 'R'=>'ARG', 'N'=>'ASN', 'D'=>'ASP', 'C'=>'CYS', 'Q'=>'GLN', 'E'=>'GLU', 'G'=>'GLY',
            'H'=>'HIS', 'I'=>'ILE', 'L'=>'LEU', 'K'=>'LYS', 'M'=>'MET', 'F'=>'PHE', 'P'=>'PRO', 'S'=>'SER',
            'T'=>'THR', 'W'=>'TRP', 'Y'=>'TYR', 'V'=>'VAL', 'B'=>'ASX', 'Z'=>'GLX', 'X'=>'UNK');

    my $workdir2="$dir/JSD_$protname";
    unless (-d "$workdir2")
    {
	`/bin/mkdir $workdir2`;
    }   
    chdir "$workdir2";
    `/bin/cp $library/$protname.pdb $workdir2/$protname\.pdb`;
    `$bin/get_fasta.pl $workdir2/$protname.pdb $workdir2/seq.txt`;
   
    if(-s "$library/JSD/$protname.blast") 
    {
        `/bin/cp $library/JSD/$protname.blast $workdir2/blast.out`;
	if(-s "$library/JSD/$protname.jsd")
	{
	    `/bin/cp $library/JSD/$protname.jsd   $workdir2/JSD.dat`;
	}
	if(-s "$library/JSD/$protname.sop")
	{
	    `/bin/cp $library/JSD/$protname.sop   $workdir2/SOP.dat`;
	}
    }
   
    if((!-s "$workdir2/JSD.dat") || (!-s "$workdir2/SOP.dat"))
    {    
	my $mod=`cat $bin/JSDmod`;
	my $tag="profile_$protname"; # unique name    
	my $jobname="$tag";
	$mod=~s/\!S\!/$protname/mg;
	$mod=~s/\!DATADIR\!/$workdir2/mg;
	$mod=~s/\!USER\!/$user/mg;

        $mod=~s/\!BLASTDIR\!/$blastdir/mg;
        $mod=~s/\!NR\!/$nr/mg;
        $mod=~s/\!BINDIR\!/$bindir/mg;
	open(FH1,">$jobname");
	print FH1 "$mod\n";
	close(FH1);
	`chmod a+x $jobname`;
	`/bin/cp $jobname $dir/.`;
	my $rst3=`./$jobname`; 

	if(!-s "$library/JSD/$protname.blast")
	{
	    `/bin/cp $workdir2/blast.out $library/JSD/$protname.blast`;
	}
	if(!-s "$library/JSD/$protname.jsd")
	{
	    `/bin/cp $workdir2/JSD.dat $library/JSD/$protname.jsd`;
	}
	if(!-s "$library/JSD/$protname.sop")
	{
	    `/bin/cp $workdir2/SOP.dat $library/JSD/$protname.sop`;
	}
    }
        
    ############# Read PDB and record residue centrality ########
    my $n   = 0; my %seqT=();  my %numT=();  my %n_atom;my  %seq1=();
    my  %x  =(); my %y =();    my %z =();
    my  %xs =(); my %ys=();    my %zs=();
    my  %xg =(); my %yg=();    my %zg=();
    my  $Lt =0; my %RSA=(); 
    open(PDB,"<$workdir2/$protname.pdb");
    while(my $line=<PDB>)
    {
        if($line=~/^ATOM/)
        {
            my $ATOM=substr($line,0,4);
            my $atom=substr($line,12,4);$atom=~s/\s//mg;$atom=~s/\d+//g;
            if($atom eq "CA")
            {    ####### for checking
                $Lt++;         
                $n_atom{$Lt}=0;
                $seq1{$Lt}=substr($line,17,3); #template aa
		$x{$Lt}=substr($line,30,8);
                $y{$Lt}=substr($line,38,8);
                $z{$Lt}=substr($line,46,8);
		$RSA{$Lt}=0.00;
                $seqT{$Lt}=$ts{substr($line,17,3)}; #only for check
                $numT{$Lt}=substr($line,22,4);
            }
            ####### record SG atoms------------->
            if($atom ne "C")
            {
                if($atom ne "CA")
                {
                    if($atom ne "N")
                    {
                        if($atom ne "O")
                        {
                            if(substr($atom,0,1) ne "H")
                            {
                                $Lt=substr($line,22,4);
                                $Lt=~s/\s//mg;
                                $n_atom{$Lt}++;
                                $xs{$Lt,$n_atom{$Lt}}=substr($line,30,8);
                                $ys{$Lt,$n_atom{$Lt}}=substr($line,38,8);
                                $zs{$Lt,$n_atom{$Lt}}=substr($line,46,8);
                            }
                        }
                    }
                }
            }
        }
    }
    close(PDB);
    for(my $k=1;$k<=$Lt;$k++)
    {
        if($n_atom{$k}< 1)
        {
            $xg{$k}=$x{$k};
            $yg{$k}=$y{$k};
            $zg{$k}=$z{$k};
        }
        else
        {
            $xg{$k}=0;
            $yg{$k}=0;
            $zg{$k}=0;
            for(my $ii=1;$ii<=$n_atom{$k};$ii++)
            {
                $xg{$k}+=$xs{$k,$ii};
                $yg{$k}+=$ys{$k,$ii};
                $zg{$k}+=$zs{$k,$ii};
            }
            $xg{$k}/=$n_atom{$k};
            $yg{$k}/=$n_atom{$k};
            $zg{$k}/=$n_atom{$k};
        }
    }
    my %Rcen=();my %RcenZ=();
    for(my $p=1;$p<=$Lt;$p++)
    {
	my $dis0=0.00;$Rcen{$p}=1.00;
	for(my $q=1;$q<=$Lt;$q++)
	{
	    if((exists $x{$p}) &&  (exists $x{$q}))
	    {
		my $dis1=sqrt(($x{$p}-$x{$q})**2+
			      ($y{$p}-$y{$q})**2+
			      ($z{$p}-$z{$q})**2);
		$dis0 +=$dis1;
	    }
	}
	$Rcen{$p}= sprintf("%7.3f",($Lt-1)/$dis0) if ($dis0 > 0);
    }
    &z_scores(\%Rcen,\$Lt,\%RcenZ);

    ####### Calculate solvent accessibility ########
    #Mean ASA taken from Protein Sci. 2003 July; 12(7): 1406-1717
   my %SA=('A'=>27, 'R'=>91, 'N'=>62, 'D'=>63, 'C'=>15, 'Q'=>75, 'E'=>81, 'G'=>26,
           'H'=>56, 'I'=>23, 'L'=>26, 'K'=>101, 'M'=>36, 'F'=>31, 'P'=>51, 'S'=>41,
           'T'=>44, 'W'=>42, 'Y'=>43, 'V'=>43, 'B'=>75, 'Z'=>75, 'X'=>26);
    my %SS=(); my %SOLV=();
    my $rst4=`$bin/stride $workdir2/$protname.pdb > $workdir2/$protname.saa 2> /dev/null`;
    open(FILE4,"<$workdir2/$protname.saa");
    while(my $line4=<FILE4>)
    {
	if($line4=~/^ASG/)
	{	    
	    my $SAS     = substr($line4,63,7);$SAS =~s/\s+//;$SAS = sprintf("%f",$SAS);	
	    my $resno   = substr($line4,10,5);$resno=~s/\s+//;$resno=sprintf("%d",$resno);	
	    my $aa      = $ts{substr($line4,5,3)};
	    $RSA{$resno}= sprintf("%5.3f",$SAS/$SA{$aa});$SOLV{$resno}= $SAS;	 
	}
    }
    close(FILE4);
    ####### Calculate residue conservation ########
    my @conserved=();my %conservedR=();my %JSDscore=();
    open(JSD,"<$workdir2/JSD.dat");
    while(my $line5=<JSD>)
    {
	if($line5=~/(\S+)\s+(\S+)/)
	{
	    my $res   =sprintf("%d",$1);
	    my $score =$2;
	    $JSDscore{$res}=$score;
	    
	    if((exists $RcenZ{$res}) && (exists $RSA{$res}))
	    {
#		print "$res $score -- $RcenZ{$res} --- $RSA{$res}\n";
		if(($score >= -0.25) && ($RcenZ{$res} > -0.2) &&($RSA{$res}>0.20))
		{
		    $conservedR{$res}=$JSDscore{$res};
		    push(@conserved,$res);
		}
	    }
	}
    }
    close(JSD);
    my %SOPscore=();
    open(JSD,"<$workdir2/SOP.dat");
    while(my $line5=<JSD>)
    {
	if($line5=~/(\S+)\s+(\S+)/)
	{
	    my $res  =sprintf("%d",$1);
	    my $score=$2;
	    $SOPscore{$res}=$score;
	    if((exists $RcenZ{$res}) && (exists $RSA{$res}))
	    {
		if(($score >= -0.25) && ($RcenZ{$res} > -0.2) && ($RSA{$res} >0.20))
		{
		    if(exists $conservedR{$res}){next;}
		    $conservedR{$res}=$SOPscore{$res};
		    push(@conserved,$res);
		}
	    }
	}
    }
    close(JSD);
    my %clusters=();my %added=();my %cluster_size=(); my %seen=();my $lastres=0;
    my @Ckeys     = reverse sort {$conservedR{$a} <=> $conservedR{$b}} keys %conservedR;
    for(my $i=0;$i<$#conserved;$i++)
    {
	my $clusterno=0;
	my $p =$Ckeys[$i];	
	if(exists $added{$i})
	{
	    $clusterno=$added{$i};
	}
	else
	{
	    $clusterno=$i;
	    $added{$i}=$i;
	    $lastres  =$i;
	    push @{$clusters{$clusterno}},sprintf("%04d",$p);
	    $seen{$p}=1;
	    $cluster_size{$clusterno}=1;
	}
	for(my $j=0;$j<$#conserved;$j++)
	{
	    my $q = $Ckeys[$j];
	    if($p eq $q){next;}
	    my @res=@{$clusters{$clusterno}};
	    my $avg_dist1=0;my $avg_dist2=0;
	    foreach my $C(@res)
	    {
		$C= sprintf("%d",$C);
		if((exists $x{$C}) && (exists $x{$q}))
		{
		    my $dis1=sqrt(($x{$C}-$x{$q})**2+
				  ($y{$C}-$y{$q})**2+
				  ($z{$C}-$z{$q})**2);
		    $avg_dist1 += $dis1;
		}
		if((exists $xg{$C}) && (exists $xg{$q}))
		{
		    
		    my $dis2=sqrt(($xg{$C}-$xg{$q})**2+
				  ($yg{$C}-$yg{$q})**2+
				  ($zg{$C}-$zg{$q})**2);		    
		    $avg_dist2 += $dis2;
		}
	    }
	    $avg_dist1= $avg_dist1/($#res+1); $avg_dist2= $avg_dist2/($#res+1);
	    if(($avg_dist1 < 15.0) || ($avg_dist2 < 15.0))
	    {
		if(exists $seen{$q}){next;}
		push @{$clusters{$clusterno}},sprintf("%04d",$q);
		$added{$q}=$clusterno;
		$cluster_size{$clusterno}++;
		$seen{$q}=1;
	    }
	}
    }
    my $clustercount=0;my @PBSITE=();
    foreach my $key (sort {$cluster_size{$b} <=> $cluster_size{$a} } keys %cluster_size)
    {
	my @res=@{$clusters{$key}};my %localS=();	
	foreach my $r(@res)
	{
	    $r=sprintf("%d",$r);
	    my $avg_dist1  =0;
	    foreach my $r2(@res)
	    {
		$r2=sprintf("%d",$r2);
		my $dis2=sqrt(($xg{$r}-$xg{$r2})**2+
			      ($yg{$r}-$yg{$r2})**2+
			      ($zg{$r}-$zg{$r2})**2);
		$avg_dist1  += $dis2;
	    }
	    $avg_dist1= $avg_dist1/($#res+1);	    
	    if($avg_dist1 < 15.0)
	    {
		$localS{$r}= $conservedR{$r}#+(1/(0.1+$avg_dist1));
	    }	    
	}	
	my @Skeys= sort {$localS{$b} <=> $localS{$a}} keys %localS;
	my $gap=1;
      redosort:;
	my @sorted2=();
	for(my $j=0;$j<$#Skeys;$j++)
	{
	    my $r=$Skeys[$j];
	    my $mindiff=$r;
	    foreach my $r2(@sorted2)
	    {
		my $diff =abs($r2-$r);
		$mindiff=$diff if($diff < $mindiff);
	    }
	    if($mindiff > $gap)
	    {
		push (@sorted2,sprintf("%04d",$r));
		$lastres=$r;
	    }	    
	}
	my @final= sort { $a <=> $b } @sorted2;
	my $r= join(",",@final);
	if($#sorted2 >1)
	{
	    $clustercount++;
	    my $acline=sprintf("$protname :ACT:BS%02d:%04d:%s",$clustercount,$#sorted2+1,$r);
	    push(@PBSITE,$acline);
	    if($clustercount > 1)
	    {
		goto finishbs;
	    }
	}
    }
  finishbs:;   
    return(@PBSITE);
}


###Cluster predicted ligands from templates
sub cluster
{
    my($dir,$Ncluster,$ED_cut,$TAN_cut,$NBS,$ED_ref,$TN_ref,$RS_ref,$LS_ref,$ID_ref,$PB_ref,$name_ref,$pocket_ref,$PC_ref,$CS_ref)=@_;
    my %EDh   = %$ED_ref;
    my %TNh   = %$TN_ref;
    my %RSh   = %$RS_ref;
    my %LSh   = %$LS_ref;
    my %bions=('MG'=>1,'ZN'=>1,'MN'=>1,'CO'=>1,'FE'=>1,'FE2'=>1,'NI'=>1,'MO'=>1,'CA'=>1,'NA'=>1,'CU1'=>1,'CU2'=>1,'CU'=>1);

    my %clusters=(); my %bool   =();
    my %Dmatrix =(); my %Cmatrix=();
    my %NEB=();

    # Parameters for very good match 
    my $good_local = 1.3; my $good_global = 0.7; my $good_id = 0.60;

    # Store all the values
    for(my $id1=0;$id1<$NBS;$id1++)
    {
        $bool{$id1}=1;
        for(my $id2=0;$id2<$NBS;$id2++)
        {
            my $dist= $EDh{$id1,$id2};
            my $tan = $TNh{$id1,$id2};
            $Dmatrix{$id1,$id2}=$dist;
            $Cmatrix{$id1,$id2}=$tan;
        }
    }
    my $i =0;
    do
    {
        my %ND=();my %NS=();my %goodM=(); my %DSsim=(); my %CSsim=();
        for(my $id1=0;$id1<$NBS;$id1++)
        {
            if($bool{$id1}==1)
            {
                $ND{$id1}=0;$NS{$id1}=0; push @{$goodM{$id1}},$id1;
                for(my $id2=0;$id2<$NBS;$id2++)
                {
                    if(($bool{$id2} == 0) || ($id1 == $id2)){next;}
                    my $dist= $EDh{$id1,$id2};
                    my $tan = $TNh{$id1,$id2};
                    if($dist <= $ED_cut)                    {
                        $ND{$id1}++;
                        push @{$goodM{$id1}},$id2;
                    }
                    if($tan >= $TAN_cut)
                    {
                        $NS{$id1}++;
                    }
                }
                if(exists $NEB{$id1}){next;}
                else{$NEB{$id1}=$ND{$id1};}
            }
            else
            {
                next;
            }
        }

        my @ranking1 = reverse sort{$ND{$a} <=> $ND{$b}} keys %ND;
        my $id1 = $ranking1[0]; my $id1_score=$RSh{$id1}; my $id1_iden=$$ID_ref{$id1};my $id1_local=$LSh{$id1};
        my @col=@{$goodM{$id1}};
        my $arrsize= $#col + 1;
#       foreach my $C (@col)
#       {
#           print "$$name_ref{$C},";
#       }
#       print "\n";         
        #### If most of the decoys already selected move to next rank             
        foreach my $C1 (@col)
        {
            $DSsim{$C1}=1;
            foreach my $C2 (@col)
            {
                if($C1 eq $C2){$DSsim{$C1}+=0;$CSsim{$C1} += 0; next;}
                $DSsim{$C1} += $EDh{$C1,$C2};
                $CSsim{$C1} += $TNh{$C1,$C2};
            }
            $DSsim{$C1} = $DSsim{$C1}/$arrsize;
            $CSsim{$C1} = $CSsim{$C1}/$arrsize;
        }
        my $mean=&Mean(\%DSsim,\@col,$arrsize);
        my $SD  =&SD(\%DSsim,\@col,\$mean,$arrsize);
        my %SS=();my $density = ($#col+1)/$NBS;
        foreach my $C1(@col)
        {

            my $y        = ($RSh{$C1} + 2/(1 + $DSsim{$C1}));
            if($#col <10){$y     = $RSh{$C1};}
            my $density  = ($NEB{$C1}/$NBS);
	    $density     = 0.1 if($density < 0.1);
            if(($density <0.35) && ($LSh{$C1} > $good_local) && ($$ID_ref{$C1} > $good_id)){$density=0.35;}
            my $x        = - $density *($y- 0.8);
            my $CscoreLB   = (2/(1+exp($x)) + 0.01)-1;
            $CscoreLB   = 0.01 if($CscoreLB < 0.01);
            $SS{$C1}     = $CscoreLB;
        }
        my @ranking2 = reverse sort{$SS{$a} <=> $SS{$b}} keys %SS;
        my $selected = $ranking2[0];

        my $xc = $$PC_ref{$selected,1}; my $yc = $$PC_ref{$selected,2}; my $zc = $$PC_ref{$selected,3}; my $in_cluster=1;
        my $id=$selected;
        my $n = $$name_ref{$id};
        foreach my $id2 (@col)
        {
            if($id eq $id2){next;}
            my $bs_sim1=&bssim($$PB_ref{$id},$$PB_ref{$id2});
            my $bs_sim2=&bssim($$PB_ref{$id2},$$PB_ref{$id});
            if(($EDh{$id,$id2} <= $ED_cut) && ($TNh{$id,$id2} >= $TAN_cut) && (($bs_sim1 >= 0.50) || ($bs_sim2 >= 0.50)))
            {
#               print "Cluster $$Ncluster Added $$name_ref{$id2} to $$name_ref{$id}  $TNh{$id,$id2} $bs_sim1 $bs_sim2\n";
                my $Lname1 = $$name_ref{$id}; $Lname1 =~s/\s+//;
                my $Lname2 = $$name_ref{$id2};$Lname2 =~s/\s+//;
                if((exists $bions{$Lname1})&&(exists $bions{$Lname2}))
                {
                    if($EDh{$id,$id2} <= 4.0)
                    {
                        $bool{$id2}=0;$i++;
                        $xc +=  $$PC_ref{$id2,1};
                        $yc +=  $$PC_ref{$id2,2};
                        $zc +=  $$PC_ref{$id2,3};
                        $in_cluster++;
                    }
                }
                elsif(exists $bions{$Lname2})
                {
                    next;
                }
                else
                {
                    $bool{$id2}=0;$i++;
                    $xc += $$PC_ref{$id2,1};
                    $yc += $$PC_ref{$id2,2};
                    $zc += $$PC_ref{$id2,3};
                    $in_cluster++;
                }
            }
            elsif(($EDh{$id,$id2}  <= $ED_cut) && ($bs_sim1 >= 0.60) && ($bs_sim2 >= 0.60))
            {
#                print "Cluster $$Ncluster Added $$name_ref{$id2} to $$name_ref{$id}  $TNh{$id,$id2} $bs_sim1 $bs_sim2\n";
                $bool{$id2}=0;$i++;#print "$id\t$id2 $i\n";
                $xc += $$PC_ref{$id2,1};
                $yc += $$PC_ref{$id2,2};
                $zc += $$PC_ref{$id2,3};
                $in_cluster++;
            }
        }
        $$Ncluster++;
        $$pocket_ref{$$Ncluster,1}=$xc/$in_cluster;
        $$pocket_ref{$$Ncluster,2}=$yc/$in_cluster;
        $$pocket_ref{$$Ncluster,3}=$zc/$in_cluster;
        $clusters{$$Ncluster}=$selected;
        $$CS_ref{$$Ncluster}=$SS{$selected};
#       print "-->$$Ncluster $selected $$pocket_ref{$$Ncluster,1}  $$pocket_ref{$$Ncluster,2} $$pocket_ref{$$Ncluster,3}\n";
        $bool{$selected}=0;
        $i++;
    }while($i<$NBS);
    return(%clusters);
}
### Identity of binding sites
sub bssim
{
    my($f1,$f2)=@_;
    my @f1bin=split(/\,/,$f1);
    my @f2bin=split(/\,/,$f2);
    my $NAB=0; my $SIM=0; my %seen_res1=();
    my $d = $#f1bin + 1;
    for(my $i=0;$i<=$#f1bin;$i++)
    {
        $seen_res1{$f1bin[$i]}=1;
    }
    for(my $i=0;$i<=$#f2bin;$i++)
    {
        $NAB++ if(exists $seen_res1{$f2bin[$i]});
    }
    $SIM = sprintf("%4.3f",($NAB/$d)) if($d > 0) ;
    return($SIM);
}

### Ligand Chemical similarity
sub tannimoto
{
        my($f1,$f2)=@_;
        my @f1bin=split(//,$f1);
        my @f2bin=split(//,$f2);
        my $bins = $#f1bin;
        $bins = $#f2bin if($#f2bin < $bins);
        my $NA=0; my $NB=0; my $NAB=0;  my $Tan=0;
        for(my $i=0;$i<=$bins;$i++)
        {
            $NA++  if($f1bin[$i] ne '0');
            $NB++  if($f2bin[$i] ne '0');
            $NAB++ if(($f1bin[$i] eq $f2bin[$i]) && ($f1bin[$i] ne '0'))
        }
        my $d = ($NA+$NB-$NAB);
        $Tan = sprintf("%4.3f",($NAB/$d)) if($d > 0) ;
        return($Tan);
}
 
#### Subroutine to mark probable binding site residues in query structure 
#### based on profile + structure aln. + sequence aln.
sub getBinding
{
  my($PBSR,$pdbfile1, $pdbfile2,$dir,$datadir,$Lch,$bindingr,$seqcons,$blast,$CS) = @_;
  my %profile=%$seqcons; my @rst=();
  my %QBS=();
  #####################Template binding residues
  my %TBS=(); my $NTBR=0;
  my @Tres=split(/\,/,$bindingr);
  foreach my $res(@Tres)
  {
      $res=sprintf("%d",$res);
      $TBS{$res}=1;     
  }
  $NTBR=@Tres;

  ##### Get Geometric Center of Template Binding site and Radius of Template binding site
  my $GCX=0.0; my $GCY=0.0; my $GCZ=0.0;
  my %X  = (); my %Y  = (); my %Z  = ();
  my $R=0;my $nres=0;
  
  open(TEMPL,"<$pdbfile2") || die "Could not find $pdbfile2\n";
  while(my $line=<TEMPL>)
  {
      if($line=~/^ATOM/){
	  if(substr($line,12,4)=~/CA/){
	      $nres++;
	      if(exists $TBS{$nres}){
		  $GCX +=substr($line,30,8);
		  $GCY +=substr($line,38,8);
		  $GCZ +=substr($line,46,8);		  
	      }
	      $X{$nres}=substr($line,30,8);
	      $Y{$nres}=substr($line,38,8);
	      $Z{$nres}=substr($line,46,8);
	  }
      }
  }
  close(TEMPL);
  $GCX=$GCX/$NTBR;
  $GCY=$GCY/$NTBR;
  $GCZ=$GCZ/$NTBR;
  my $pcut=0;
  my $nbres1=0;
  for(my $i=1;$i<=$nres;$i++)
  {
      if(exists $TBS{$i})
      {
	  $nbres1++;
	  my $dis=sqrt(($X{$i}-$GCX)**2+
		       ($Y{$i}-$GCY)**2+
		       ($Z{$i}-$GCZ)**2);
	  if($dis > $R)
	  {
	      $R=sprintf("%6.3f",$dis);
	  }
      }
  }
  ##########Superpose Query model and Template & get Rotation and Translation matrix
  my @rot=();my @tra=();
  @rst=(); my $TM =0;
  @rst=`$dir/TMalign $pdbfile1 $pdbfile2`;
  my @aln1=();my @aln2=();
  my @index=("14","15","16");
  for(my $i=0;$i<$#rst;$i++)
  {
     if($rst[$i]=~/\s+1\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/)
     {
      $tra[0]=$1;
      $rot[0][0]=$2;
      $rot[0][1]=$3;
      $rot[0][2]=$4;
     }
     if($rst[$i]=~/\s+2\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/){
      $tra[1]=$1;
      $rot[1][0]=$2;
      $rot[1][1]=$3;
      $rot[1][2]=$4;
     }
     if($rst[$i]=~/\s+3\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/){
      $tra[2]=$1;
      $rot[2][0]=$2;
      $rot[2][1]=$3;
      $rot[2][2]=$4;
     }
     if($rst[$i]=~/TM-score\s*=(\S+),/)
     {
        $TM = $1;
     }
     if($rst[$i]=~/^\(/)
     {
        @aln1=split(//,$rst[$i+1]);
        @aln2=split(//,$rst[$i+3]);
     }
  }
  if($TM >= 0.45)
  {
      $pcut =0.25;
  }
  my $Qindex=0;my $Tindex=0;
  for(my $pos=0; $pos<$#aln1; $pos++)
  {
      if ($aln1[$pos] ne "-" && $aln2[$pos] ne "-")
      {
          if(exists $TBS{$Tindex}){$QBS{$Qindex}=1;}
      }
      if ($aln1[$pos] ne "-")
      {
          $Qindex++;
      }
      if ($aln2[$pos] ne "-")
      {
          $Tindex++;
      }
  }
  ##########Superpose Query model and get probable binding site region in the model######################
  for (my $i=1;$i<=$Lch;$i++)
  {
      $$PBSR{$i}=-1;
      if($profile{$i} >=$pcut)
      {
	 $$PBSR{$i}=$profile{$i}; 
      }
      if(($CS < 0.4) && ($profile{$i} > 0.0))
      {
	  $$PBSR{$i}=0.31;
      }
  }

  my $qres = -1000;my $start = 1;my $last_res_id=''; 
  open(QUE,"<$pdbfile1");
  while(my $line=<QUE>)
  {
      if($line=~/^ATOM/)
      {
	  if ($qres == -1000)
	  { # start with the starting number in the chain	     
	      $qres = $start;
	      $last_res_id = substr($line,22,5);
	  }
	  my $res_id = substr($line,22,5);
	  if ($res_id ne $last_res_id)
	  {
	      $qres++;
	      $last_res_id = $res_id;
	  }
	  my $xc=substr($line,30,8);
	  my $yc=substr($line,38,8);
	  my $zc=substr($line,46,8);
	   
	  my $nx=$rot[0][0]*$xc+$rot[0][1]*$yc+$rot[0][2]*$zc+$tra[0];
	  my $ny=$rot[1][0]*$xc+$rot[1][1]*$yc+$rot[1][2]*$zc+$tra[1];
	  my $nz=$rot[2][0]*$xc+$rot[2][1]*$yc+$rot[2][2]*$zc+$tra[2];
	  my $dis=sqrt(($nx-$GCX)**2+
		       ($ny-$GCY)**2+
		       ($nz-$GCZ)**2);
	  my $cut = 2;        
          my $nbres2= 0;
	  if($dis <= $R+$cut)
	  {	 
	      if($profile{$qres} >=$pcut)
	      {
		  $$PBSR{$qres}=$profile{$qres};
		  $nbres2++;
	      }
	      elsif(exists $QBS{$qres})
	      {
		  $$PBSR{$qres}=1;
	      }
	  }
      }
  }
  close(QUE);
  system("./get_fasta.pl $pdbfile2 template.seq");
  my @blast_out=`$blast/bin/bl2seq -i protein.seq -j template.seq  -p blastp > blast_align.out`;
  open(BLA,"<blast_align.out");
  while(my $bline=<BLA>)
  {  
      if($bline=~/Expect =\s*(\S+)/)
      {
	  <BLA>;
	  <BLA>;
	  while($bline=<BLA>)
	  {
	      if($bline=~/Query\:\s*(\d+)\s+(\S+)\s+(\S+)/)
	      {
		  my $i1=$1;
		  my $seq1=$2;
		  <BLA>;
		  <BLA>=~/Sbjct\:\s*(\S+)\s+(\S+)\s+(\S+)/;
		  my $seq2=$2;
		  my $j1=$1;
		  <BLA>;
		  ###
		  my $L=length $seq1;
		  my $m1=$i1-1;
		  my $n1=$j1-1;
		  for(my $i=1;$i<=$L;$i++)
		  {
		      my $q1=substr($seq1,$i-1,1);
		      my $q2=substr($seq2,$i-1,1);
		      $m1++ if($q1 ne '-');
		      $n1++ if($q2 ne '-');
		      if($q1 ne '-' && $q2 ne '-')
		      {			 
			  if(exists $TBS{$n1})
			  {
			      $$PBSR{$m1}=1;
			  }
		      }
		  }		  
	      }
	      else
	      {
		  goto pos11;
	      }	      
	  }
	pos11:;
      }
  }
  close(BLA);
  return;
}

sub runJSD
{
    (my $protname, my $dir1, my $user, my $blast, my $NR, my $bin)=@_;
    my $mod=`cat $bin/JSDmod`;
    my $tag="profile_$protname"; # unique name    
    my $jobname="$tag";
    $mod=~s/\!S\!/$protname/mg;
    $mod=~s/\!DATADIR\!/$dir1/mg;
    $mod=~s/\!USER\!/$user/mg;
    $mod=~s/\!BLASTDIR\!/$blast/mg;
    $mod=~s/\!NR\!/$NR/mg;
    $mod=~s/\!BINDIR\!/$bin/mg;
    open(FH1,">$jobname");
    print FH1 "$mod\n";
    close(FH1);
    `chmod a+x $jobname`;
    `/bin/cp $jobname $dir1/.`;
    `./$jobname`;
    return;
}

sub makePRF
{
    (my $protname, my $dir1, my $user, my $blast, my $NR, my $bin)=@_;
    my $mod=`cat $bin/PRFmod`;
    my $tag="PRF_$protname"; # unique name    
    my $jobname="$tag";
    $mod=~s/\!S\!/$protname/mg;
    $mod=~s/\!DATADIR\!/$dir1/mg;
    $mod=~s/\!USER\!/$user/mg;
    $mod=~s/\!BLASTDIR\!/$blast/mg;
    $mod=~s/\!NR\!/$NR/mg;
    $mod=~s/\!BINDIR\!/$bin/mg;
    open(FH1,">$jobname");
    print FH1 "$mod\n";
    close(FH1);
    `chmod a+x $jobname`;
    `/bin/cp $jobname $dir1/.`;
    `./$jobname`;
    return;
}

### Subroutine for calculating Mean
sub Mean 
{
	my($hash_ref,$arr_ref,$n)=@_;   
	my %x = %$hash_ref; 	
	my @arr=@$arr_ref;
	my $sum=0;my $i=0;
	foreach my $a(@arr)
	{
	    $sum += $x{$a};	    
	}

	my $m=$sum/($n);
	return $m;
}

### Subroutine for calculating Standard-deviation
sub SD 
{
    my($hash_ref,$arr_ref,$mean_ref,$n)=@_;
    my %x = %$hash_ref; 
    my @arr=@$arr_ref;
    my $i=0; my $sum=0;
    foreach my $a(@arr)
    {    
	$sum=$sum+($x{$a}-$$mean_ref)*($x{$a}-$$mean_ref);
    }
    my $Std=sqrt($sum);
    $Std=sqrt($sum/($n-1))if($n >1);    
    return $Std;
}
### Subroutine for calculating run-time
sub runhour
{
    my($time1)=@_;
    my $time2=time();
    my $time=($time2-$time1)/3600;
    return($time);
}

sub z_scores
{# This subroutine calculates Z-scores for a set of scores (NOTE: hash index starts from 1)   
     my($score_ref,$L,$z_score_ref)=@_;
     my %r_score  =%$score_ref;
     my $Lch   =$$L;
     my $z1_a  = 0; my $z1_a2=0; my %zscore=();
     for(my $i=1;$i<=$Lch;$i++)
     {
         $z1_a  += $r_score{$i};
         $z1_a2 += $r_score{$i}**2;
     }
     $z1_a/=$Lch;
     $z1_a2/=$Lch;
     my $z1_sd=sqrt($z1_a2-$z1_a**2);
     for(my $i=1;$i<=$Lch;$i++)
     {
          $$z_score_ref{$i}=($r_score{$i}-$z1_a)/$z1_sd;
     }
}
   

### Subroutine for calculating C-score of structure modeling
sub getCscore
{ 
    my($dir,$cscore_ref)=@_;
    my %NAME=(1=>"QQQ",2=>"IIIe",3=>"WWW",4=>"VVV",5=>"BBB",6=>"GGGd",7=>"IIIj",8=>"AAA",9=>"EEE",10=>"UUU",11=>"CCC",12=>"NNNd",13=>"OOO");
    
    my @inits=("QQQ","IIIe","WWW","VVV","BBB","GGGd","IIIj","AAA","EEE","UUU","CCC","NNNd","OOO");
						 
    my %zscore0=('QQQ'=>5.8,'IIIe'=>9.5,'WWW'=>6,'VVV'=>7.0,'BBB'=>3.2,'GGGd'=>9.3,'IIIj'=>11.5,'AAA'=>19,'EEE'=>7,'UUU'=>7.0,'CCC'=>15,'NNNd'=>7.0,'OOO'=>25,'III'=>40);	 	    
    my $init="init.dat";

    if(-s "$dir/init2.dat"){
        $init="init2.dat";
    }
    else{
        $init="init.dat";
    }


    my $M=13; my $k =0; my $z_sum =0; 
    for(my $i=1;$i<=$M;$i++){
	open(FH,"$dir/$init");
	<FH>=~/(\d+)/;
	my $nt=$1;
	my %zscore=();
	my $n=0;my %L=();
	for(my $j=1;$j<=$nt;$j++){
	    <FH>=~/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/;
	    $L{$j}=$1;
	    my $from=$5;
	    if($from eq "$NAME{$i}"){
		$n++;
		$zscore{$n}=$2;
                if($zscore{$n} < 0){$zscore{$n}=0;}
	    }
	    for(my $o=1;$o<=$L{$j};$o++){
		<FH>;
	    }
	    <FH>;
	}
      pos2:;
	close(FH);
	
	if($n > 0){
	    my @zscore_keys=sort{$zscore{$b}<=>$zscore{$a}} keys %zscore;
	    $k++;
	    my $z_max=$zscore{$zscore_keys[0]};
	    if($NAME{$i} eq "III"){
		$z_max=$zscore{1};
	    }
	    my $z_nor=$z_max/$zscore0{$NAME{$i}};
	    $z_sum+=$z_nor;
	}
    }
    my $zn=$z_sum/$k;
    my $rst="$dir/rst.dat";
    if(-s "$dir/rst.dat"){$rst="$dir/rst.dat";}
    elsif(-s "$dir/cluster/rst.dat"){$rst="$dir/cluster/rst.dat";}
    elsif(-s "$dir/cluster_new/rst.dat"){$rst="$dir/cluster_new/rst.dat";}
    elsif(-s "$dir/CLUSTER/rst.dat"){$rst="$dir/CLUSTER/rst.dat";}
    elsif(-s "$dir/cluster2/rst.dat"){$rst="$dir/cluster2/rst.dat";}
    else{print "Error $rst was not found\n";exit();}
    
    my $nc=0;my $nstr=0;
    open(FH,"<$rst"); 
    while(my $line=<FH>){
	my $Lch=$1  if($line=~/Modeling Length:\s+(\d+)/);
	if($line=~/Number of structure in use\=\s*(\S+)/){$nstr=$1;}
	if($line=~/Number of clusters\:\s+(\S+)/){
	    $nc=$1; $nc=5    if($nc>5);
	   
	}	
	my $Rcin='';
	if($line=~/C-----/){	   
	    for(my $i=1;$i<=$nc;$i++){
		my $line2=<FH>;
		$line2=~/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/;
		my $Nin=$2;  		
		my $cut=0.8;
		if($i==1){
		    $Rcin=$4;
		    $cut=0.8;
		    $Rcin=$cut if($Rcin<$cut);
		}	
		my $density=$Nin/$nstr/$Rcin;  
		$$cscore_ref{$i}=sprintf("%6.3f",log($zn*$density));               
	    }
	}
    }
    close(FH);
}
